from datetime import datetime, timedelta, timezone
import pytz
import time
import calendar

def timer_decorator(func): 
    '''
    @description: 计时装饰器，测量函数执行时间
    @param: func: 被装饰的函数
    @return: wrapper: 装饰器函数
    @usage:
    @timer_decorator
    def test():
    print("Hello, world!")
    time.sleep(1)
    print("Goodbye, world!")
    test()
    # 输出：
    # 函数 test 执行耗时: 1.000000 秒       
    '''   
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print(f"函数 {func.__name__} 执行耗时: {end_time - start_time:.6f} 秒")
        return result
    return wrapper


def convert_julian_regular_date(julian_date):
    # 将Julian日期转换为正常的日期
    year = int(str(julian_date)[:4])  # 提取年份
    day_of_year = int(str(julian_date)[4:])  # 提取一年中的第几天
    # 创建日期对象
    date = datetime(year=year, month=1, day=1)  # 设置为当年的第一天
    date += timedelta(days=day_of_year - 1)  # 增加相应天数
    print("转换后的日期:", date.strftime("%Y-%m-%d"))
    return date

def get_UTC_time(start_date,days=0):
    '''

    Args:
        start_date: str, 如'2000-01-01 00:00:00'
        days: float,如7670.0625

    Returns:返回utc时间
    '''
    # 定义起始时间
    start_time_utc=None
    if isinstance(start_date,datetime):
        start_time_utc=start_date.replace(tzinfo=timezone.utc)        
    else:
       start_time_utc = datetime.strptime(start_date.astype(str).replace("000000000",'00'), '%Y-%m-%dT%H:%M:%S.%f').replace(tzinfo=timezone.utc)
    # 定义时间增量（7670.0625 天）
    days_increment = timedelta(days=days)
    # 计算当前时间
    current_time = start_time_utc + days_increment
    return current_time


def get_Beijing_time_from_timezone(date_with_timezone,days=0):
    '''
    Args:
    @ date_with_timezone: 如'2020-12-31T01:30:00.000000000'
       
    Returns:返回北京时间
    '''
    # 将UTC时间转换为北京时间
    utc_timezone = pytz.timezone('UTC')
    beijing_timezone = pytz.timezone('Asia/Shanghai')

    # 将带有时区信息的时间字符串转换为 datetime 对象
    datetime_obj = datetime.strptime(date_with_timezone.astype(str).replace("000000000",'00'), '%Y-%m-%dT%H:%M:%S.%f')
    utc_datetime = utc_timezone.localize(datetime_obj)
    beijing_datetime = utc_datetime.astimezone(beijing_timezone)
    beijing_datetime=beijing_datetime+ timedelta(days=days)
    return beijing_datetime.astimezone(tz=None).replace(tzinfo=None)
    # 将 datetime 对象转换为常规时间字符串
    #formatted_time = datetime_obj.strftime('%Y-%m-%d %H:%M:%S')


def get_Beijing_time(start_date, days):
    '''
    Args:
        start_date: datetime
        days: float,如7670.0625
    Returns:返回beijing时间
    '''
    days_increment = timedelta(days=days)
    # 计算当前时间
    current_time = start_date + days_increment
    current_time = current_time.astimezone(tz=None).replace(tzinfo=None)  # 将时间转换为本地时区
    # print(f"当前时间：{current_time}")
    return current_time

def get_Beijing_time_from_UTC(start_date, days):
    '''
    Args:
        start_date: str, 如'2000-01-01 00:00:00'
        days: float,如7670.0625
    Returns:返回beijing时间
    '''
    current_time = get_UTC_time(start_date, days)
    current_time = current_time.astimezone(tz=None).replace(tzinfo=None)  # 将时间转换为本地时区
    print(f"当前时间：{current_time}")
    return current_time

def get_Beijing_time_from_UTC(utc_time):
    '''
    Args:
        start_date: datetime, 如2020-12-31 01:30:00+00:00
    Returns:返回beijing时间
    '''
    current_time = utc_time.astimezone(tz=None).replace(tzinfo=None)  # 将时间转换为本地时区
    print(f"当前时间：{current_time}")
    return current_time

def get_closest_date_index(datetime_list,given_time):
    '''
    Args:
        datetime_list:datetime 对象的集合
        given_time:给定的时间
    Returns:
    '''

    # 查找集合中最接近给定时间的元素的索引
    closest_index = min(range(len(datetime_list)), key=lambda i: abs(datetime_list[i] - given_time))
    return closest_index

def get_closest_hour_index(datetime_list,given_time):
    '''
    Args:
        datetime_list:datetime 对象的集合
        given_time:给定的时间
    Returns:
    '''
    # 查找集合中最接近给定时间的元素的索引
    closest_day_index = min(range(len(datetime_list)), key=lambda i: abs(datetime_list[i] - given_time))
    closest_date=datetime_list[closest_day_index]
    hours=[]
    for date in datetime_list:
        if date.year==closest_date.year and date.month==closest_date.month and date.day==closest_date.day:
            hours.append(date.hour)
    closest_hour_index = min(range(len(hours)), key=lambda i: abs(hours[i] - given_time.hour))
    if given_time.tzinfo is not None:
        closest_hour_date=datetime(closest_date.year,closest_date.month,closest_date.day,hours[closest_hour_index],closest_date.minute,tzinfo=given_time.tzinfo)
    else:
        closest_hour_date=datetime(closest_date.year,closest_date.month,closest_date.day,hours[closest_hour_index],closest_date.minute)    
    closest_index=datetime_list.index(closest_hour_date)
    return closest_index



def get_hour_of_year(date):
    '''
    @description: 获取指定日期属于当年第x个小时
    @param: date: 日期
    @return: 获取指定日期属于当年第x个小时
    @usage:
    '''  
    # 获取当年的第一天
    start_of_year = datetime(date.year, 1, 1)    
    # 计算给定日期与当年第一天之间的时间差
    time_difference = date - start_of_year    
    # 将时间差转换为小时数
    hours_of_year = time_difference.total_seconds() / 3600    
    # 加上当前日期的小时数
    hours_of_year += date.hour    
    return int(hours_of_year)


def get_days_in_month(month, year):
    '''
    @description: 获取指定月份的天数
    @param: month: 月份
    @param: year: 年份
    @return: 月份的天数
    @usage:
    print(get_days_in_month(2, 2021))  # 29
    print(get_days_in_month(4, 2021))  # 30
    '''
    return calendar.monthrange(year, month)[1]


class FactorInfo:
    def __init__(self):
        self.Region = ""
        self.Pollutant = ""
        self.Source = ""
        self.Min = 0.0
        self.Max = 1.2
        self.Limit = 0.2
        self.BaseEmissionRatio = 0.0
        self.Index = 0


import numpy as np
import matplotlib.pyplot as plt

#对数据进行归一化
def changex(x, fiSize):
    xmin = np.min(x)
    x = x - xmin
    x = x / np.max(x) * fiSize
    return x
#拟合直线，绘制散点密度图
'''
data:ndarray 数组，不能是数据框，数据框的话，需要调用to_numpy()转为数组
titleName:标题
xName：x轴标题
yName：y轴标题
fiSize：数据归一化大小（如100，即将数据归一化为0~100之间的数）
'''
def draw_density_plot(x, y, titleName, xName, yName, fiSize=100,showMNE=True):
    # x = data[:, 0]  # RSM
    # y = data[:, 1]  # CMAQ

    denSta = np.zeros((fiSize, fiSize))#这里模拟了x*y=fiSize*fiSize的格子，待会要用来放统计每个格子里有几个点的结果
    xlength = len(x)
    xdenSta = changex(x, fiSize)
    ydenSta = changex(y, fiSize)

    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        denSta[m, n] = denSta[m, n] + 1

    z = np.zeros((x.shape[0], 1))
    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        z[count, 0] = denSta[m, n]

    int_denSta=denSta.flatten().astype(int)
    #tabulateDensta = np.bincount(int_denSta)
    #计算每个元素出现的频率，unique_elements为元素，tabulateDensta为元素对应下的出现频率
    unique_elements, tabulateDensta= np.unique(int_denSta, return_counts=True)
    #准备装累积百分比列表的矩阵
    percent = np.zeros(tabulateDensta.shape)
    percent[0] = 0 #把0的百分比归0

    sumPoint = x.shape[0]
    #unique_tabulateDensta = np.unique(int_denSta)

    #装累积百分比查询表
    for count in range(1, tabulateDensta.shape[0]):
        percent[count] = percent[count - 1] + (unique_elements[count] * tabulateDensta[count] / sumPoint)
#傻瓜遍历找到原z值对应的百分比，更新z
    for count in range(z.shape[0]):
        for coutFindPercent in range(unique_elements.shape[0]):
            if z[count, 0] == unique_elements[coutFindPercent]:
                z[count, 0] = percent[coutFindPercent]
                break

    fig = plt.figure(figsize=(11.5, 9.8))
    plt.scatter(y, x, s=26, c=z, marker='s', cmap='jet')#这里调换了一下顺序，使得x轴表示的是CMAQ，y轴表示的是RSM
    #拟合直线 + 拟合评价
    minAll = np.floor(np.min([np.min(x), np.min(y)]) / 20) * 20
    maxAll = (np.floor(np.max([np.max(x), np.max(y)]) / 20) + 1) * 20
    #2点确定一条直线 （参考线）
    plt.plot([minAll, maxAll], [minAll, maxAll], 'b-.')
    if np.min([np.min(x), np.min(y)]) < 0:
        plt.axhline(y=0, color='k', linestyle='--')
        plt.axvline(x=0, color='k', linestyle='--')
    #2点确定一条直线 (拟合线)
    PolyfitResult = np.polyfit(x, y, 1)
    plt.plot([minAll, maxAll], [PolyfitResult[0] * minAll + PolyfitResult[1], PolyfitResult[0] * maxAll + PolyfitResult[1]], 'r-')

    #评价拟合质量
    N = x.shape[0]
    R2 = 1 - np.sum((PolyfitResult[0] * x + PolyfitResult[1] - y) ** 2) / np.sum((y - np.mean(y)) ** 2)

    Ec = np.abs(x - y) / np.abs(y)
    ETc = Ec[Ec <= 1000]
    Fc = np.abs(x - y) / np.abs(x + y)
    FTc = Fc[Fc <= 1000]

    MNE = np.sum(ETc) / (N) * 100
    MFE = np.sum(FTc) / (N) * 100
    RMSE = np.sqrt(np.mean((y - x) ** 2))#计算均方根误差
    MAE = np.mean(np.abs(y - x))

    MaxNE = np.sort(ETc)* 100
    MaxNE_95 =np.percentile(MaxNE, 95) #percentileofscore(MaxNE, 95)
    MaxFE = np.sort(FTc)* 100
    MaxFE_95 =np.percentile(MaxFE, 95)# percentileofscore(MaxFE, 95)

    plt.xlabel(xName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    plt.ylabel(yName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    plt.title(titleName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    str = ['Y = {:0.4f} + {:0.4f}X'.format(PolyfitResult[1], PolyfitResult[0]),
           'R^2 = {:0.4f}'.format(R2),
           'NME = {:0.2f}%'.format(MNE),
           '95th MaxNE = {:0.2f}%'.format(MaxNE_95),
           'RMSE = {:0.4f}'.format(RMSE),
           'MAE = {:0.4f}'.format(MAE)]
    if not showMNE:
        str = ['Y = {:0.4f} + {:0.4f}X'.format(PolyfitResult[1], PolyfitResult[0]),
               'R^2 = {:0.4f}'.format(R2),
               'RMSE = {:0.4f}'.format(RMSE),
               'MAE = {:0.4f}'.format(MAE)]
    plt.text(0.04, 0.95, '\n'.join(str), fontname='Times New Roman', verticalalignment='top', horizontalalignment='left', fontweight='bold',
             fontsize=10,transform=plt.gca().transAxes)

    plt.xlim(minAll, maxAll)#设置x轴的数据显示范围
    plt.ylim(minAll, maxAll)#设置y轴的数据显示范围
    plt.xticks(np.arange(minAll, maxAll, 10))
    plt.yticks(np.arange(minAll, maxAll, 10))
    plt.tight_layout()

    #plt.colorbar(label='Percent of data point', fontweight='bold', fontsize=35)
    cbar = plt.colorbar(label='Percent of data point', orientation='vertical')
    # Set font properties for the colorbar label
    #cbar.ax.yaxis.label.set_font_properties({'weight': 'bold', 'size': 35})
    cbar.ax.yaxis.label.set_family('Times New Roman')
    cbar.ax.yaxis.label.set_fontweight('bold')
    cbar.ax.yaxis.label.set_fontsize(14)
    # Set the locator and formatter for the colorbar
    from matplotlib.ticker import MultipleLocator, FormatStrFormatter
    # 方法1：Set the ticks for the colorbar
    #ticks=np.arange(0, 1.05, 0.1)# Set the interval between ticks
    #cbar.set_ticks(ticks)  # Set the desired ticks
    cbar.locator = MultipleLocator(0.1)  # Set the interval between ticks
    cbar.update_ticks()
    #方法2
    '''
    cbar.locator = MultipleLocator(0.1)  # Set the interval between ticks
    cbar.update_ticks()
    # Add 0 as a tick
    ticks = cbar.get_ticks().tolist()
    if 0 not in ticks:
        ticks.append(0)
        cbar.set_ticks(sorted(ticks))
    '''
    plt.box(True)
    #plt.axis('normal')
    plt.axis('auto')
    #plt.show()
    #plt.pause(2)#加上plt.pause(0.001)，这样程序就会暂停执行一段极短的时间，然后继续执行后面的代码，而绘图界面仍然保持打开状态。
    # 关闭窗口
    #plt.gcf().canvas.mpl_disconnect(plt.gcf().canvas.manager.key_press_handler_id)
    return (RMSE, MAE, fig, MNE, MaxNE_95, R2)




'''
Author: Devin
Date: 2024-06-18 21:30:32
LastEditors: Devin
LastEditTime: 2024-06-18 22:20:26
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
from .date_helper import *
from .earth_helper import *
from .file_helper import *
from .log_helper import *
from .interpolation_helper import *
from .map_helper import *
from .netcdf_helper import *
from .PaneLayout import *
from .projection_helper import *
from .scatter_plot_helper import *
from  .sqlite_helper import *



import netCDF4 as nc
import pandas as pd
import numpy as np
import xarray as xr
from datetime import datetime

def extract_variables_data(input_file, output_file, variables,dimension_name=None):
    '''
    @description: Extract variables data from input NetCDF file and save them to output NetCDF file.if dimension_name is not None, the specified dimension will be reduced to size 1.
    @param input_file: Input NetCDF file path.
    @param output_file: Output NetCDF file path.
    @param variables: List of variables to extract.
    @param dimension_name: Name of the dimension to reduce.
    '''
    # Open the input NetCDF file
    with nc.Dataset(input_file, 'r') as src:
        # Create the output NetCDF file
        with nc.Dataset(output_file, 'w') as dst:
            # Copy global attributes
            dst.setncatts(src.__dict__)            
            # Copy dimensions, except bottom_top which will be reduced to size 1
            for name, dimension in src.dimensions.items():
                if dimension_name is not None and name == dimension_name:                   
                    dst.createDimension(name, 1)
                else:
                    dst.createDimension(name, len(dimension) if not dimension.isunlimited() else None)               
            # Copy variables
            for name, variable in src.variables.items():
                if name in variables:
                    # Create the variable in the destination file
                    out_var = dst.createVariable(name, variable.datatype, variable.dimensions)
                    out_var.setncatts(variable.__dict__)                    
                    # Copy the data, slicing the first layer for bottom_top dimension
                    data = variable[:]
                    if dimension_name in variable.dimensions:
                        if dimension_name is not None:
                            index = variable.dimensions.index(dimension_name)
                            data = np.take(data, indices=0, axis=index)#indices=0: 获取索引为0的元素，即返回数组中索引为0的元素或子数组。                        
                    out_var[:] = data


def process_24hours_data_to_bj_time(date,output_folder,source_folder,layer_name,variables=[],input_file_prefix="",input_file_suffix="nc",output_file_suffix="CH4_ANFlux.bj.nc"): 
    '''
    @description: 将24小时数据(1个含24小时文件)合并到北京时间，并提取第一层数据
    @param {date: datetime.datetime, 日期}
    @param {output_folder: str, 输出文件夹}
    @param {source_folder: str, 数据源文件夹}
    @param {layer_name: str, 层名称}
    @param {variables: list, 需要提取的变量}
    @param {input_file_prefix: str, 文件前缀}
    @param {input_file_suffix: str, 文件后缀}
    @param {output_file_suffix: str, 输出文件后缀}
    @return: None
    '''
    import os
    import shutil
    from datetime import timedelta
    output_file = os.path.join(output_folder, f"{date.strftime('%Y%m%d%H')}.all_layers.{output_file_suffix}.nc") if len(variables) > 0 else os.path.join(output_folder, f"{date.strftime('%Y%m%d%H')}.{output_file_suffix}.nc")    
    current_date_file = os.path.join(source_folder, f"{input_file_prefix}.{date.strftime('%Y%m%d')}.{input_file_suffix}")
    if not  os.path.exists(current_date_file):
        print(f"File {current_date_file} not exists, skip it.")
        return
    shutil.copy(current_date_file,output_file)
    if date.strftime('%Y%m%d') != '20220101':#第一天不做处理
        previous_date = date - timedelta(days=1)
        previous_date_file=os.path.join(source_folder, f"{input_file_prefix}.{previous_date.strftime('%Y%m%d')}.{input_file_suffix}") 
        if not os.path.exists(previous_date_file):
            print(f"File {previous_date_file} not exists, skip it.{current_date_file} failed to process.")
            return       
        # shutil.copy(current_date_file,output_file)
        split_hours=8
        with nc.Dataset(output_file, 'r+') as src_dataset:
            for variable_name, variable in src_dataset.variables.items():
                if variable.dtype == 'S1':  # 字符串类型
                    continue
                previous_8h_data = nc.Dataset(previous_date_file, 'r')[variable_name][-split_hours:, :, :, :]  # 后一天的数据 
                current_16h_data =nc.Dataset(current_date_file, 'r')[variable_name][:(24-split_hours), :, :, :] # 当前文件的数据
                # 创建新的数据数组，长度为24小时
                new_data = np.zeros((24, *current_16h_data.shape[1:])) #*表示解包，即将一个可迭代对象（如列表、元组等）解包成单独的元素           
                # 填补前9小时数据
                new_data[:split_hours, :, :, :] = previous_8h_data            
                # 填补后面的15小时数据
                new_data[split_hours:, :, :, :] = current_16h_data            
                # 写回当前文件   
                variable[:] = new_data  
    if os.path.exists(output_file) and len(variables) > 0:     
        extract_variables_data(output_file, os.path.join(output_folder, f"{date.strftime('%Y%m%d%H')}.{output_file_suffix}"), variables,dimension_name=layer_name)
        os.remove(output_file)

def process_24files_data_to_bj_time(date,time_steps, output_folder,source_folder,layer_name,variables=[],input_file_prefix="",input_file_suffix="nc",input_date_str_format='%Y_%m_%d_%H',
                 output_file_suffix="ch4.conc.bj.nc", skip_file_length_validation=False):
    '''
    @description: 将24小时数据(24个小时文件)合并到北京时间，并提取第一层数据
    @param {date: datetime.datetime, 日期}
    @param {time_steps: list, 时间步长}
    @param {output_folder: str, 输出文件夹}
    @param {source_folder: str, 数据源文件夹}
    @param {layer_name: str, 层名称}
    @param {variables: list, 需要提取的变量}
    @param {input_file_prefix: str, 文件前缀}
    @param {input_file_suffix: str, 文件后缀}
    @param {input_date_str_format: str, 输入文件日期格式}
    @param {output_file_suffix: str, 输出文件后缀}
    @param {skip_file_length_validation: bool, 是否跳过文件长度验证}
    @return: None
    '''
    import os
    import subprocess
    from datetime import timedelta
    nccopy_command = 'nccopy'
    
    # 找到所有文件
    existing_files = []
    for time_step in time_steps:
        current_date = date + timedelta(hours=int(time_step))        
        file = os.path.join(source_folder, f"{input_file_prefix}{current_date.strftime(input_date_str_format)}.{input_file_suffix}")
        if os.path.exists(file) and os.path.getsize(file) > 0:            
            existing_files.append(file)
        else:
            print(f"File {file} not found or empty.")
    if not skip_file_length_validation:    
        if  len(existing_files) != 24:
            print(f"Not all 24 files found for date {date.strftime(input_date_str_format)}")
            if date.strftime('%Y-%m-%d') != '2022-01-01': 
                return
    
    existing_extracted_files=[]
    for source_file in existing_files:
        base_dir=os.path.dirname(source_file) 
        target_file_name=os.path.basename(source_file)     
        target_file = os.path.join(base_dir, f'{"extract"}.{target_file_name}')
        variable_names = ','.join(variables)
        command = [nccopy_command, '-V', variable_names, source_file, target_file]
        subprocess.run(command)
        if os.path.exists(target_file):
            target_file_1st_layer = os.path.join(output_folder, f'{target_file_name}.1layer')
            extract_variables_data(target_file, target_file_1st_layer, variables,dimension_name=layer_name)
            if os.path.exists(target_file_1st_layer):
                existing_extracted_files.append(target_file_1st_layer)
                os.remove(target_file)           
    # 打开多个文件并合并它们
    datasets = [xr.open_dataset(file) for file in existing_extracted_files]
    # 合并数据集
    combined_dataset = xr.concat(datasets, dim='Time')
    output_file = os.path.join(output_folder, f"{date.strftime('%Y%m%d%H')}.{output_file_suffix}")
    
    # 将合并后的数据集保存为一天 24 小时的 NetCDF 文件
    combined_dataset.to_netcdf(output_file)    
    # 关闭数据集
    combined_dataset.close()
    for file in existing_extracted_files:
        os.remove(file)        

     
def readNetCDF(nc_path, pollutant,timeStepIndex=0,layerStepIndex=0):
    '''
    :param nc_path: netcdf文件路径
    :param pollutant: 读取文件中变量
    :param timeStepIndex: 时间维度索引
    :param layerStepIndex: 层数维度索引
    :return: 读取 NetCDF 文件中的变量返回行列的数据框
    '''
    nc_data = nc.Dataset(nc_path)
    nc_data = nc_data.variables[pollutant][:][timeStepIndex,layerStepIndex]
    pd_data = pd.DataFrame(nc_data)
    # 关闭 NetCDF 文件
    #nc_data.close()
    return pd_data

def readNetCDFAsFlatten(nc_path, pollutant,timeStepIndex=0,layerStepIndex=0):
    '''
    :param nc_path: netcdf文件路径
    :param pollutant: 读取文件中变量
    :param timeStepIndex: 时间维度索引
    :param layerStepIndex: 层数维度索引
    :return: 读取 NetCDF 文件中的变量并将其展平成一维数组
    '''
    nc_data = nc.Dataset(nc_path)
    nc_data = nc_data.variables[pollutant][:][timeStepIndex, layerStepIndex]
    # 将多维数组展平成一维数组
    data_flattened = nc_data.flatten()
    # 关闭 NetCDF 文件
    #nc_data.close()
    return data_flattened

def createNETCDF(template_path,output_path,input_values,template_variable_index=0):
    '''
    :param template_path: netcdf模板文件路径
    :param output_path: 要生成的netcdf文件路径
    :param input_values:
    :param template_variable_index:
    :return:
    '''
    tempalate_dataset = nc.Dataset(template_path)  # 读取该ACONC文件基本信息
    col = tempalate_dataset.variables[template_variable_index].shape[0]  # 获取变量的列数量
    row = tempalate_dataset.variables[template_variable_index].shape[1]  # 获取变量的行数量
    variable_data = np.zeros((col, row))  # 创建与需要修改的变量同维度的全0矩阵
    k = 0  # 为保证CSV顺序与矩阵顺序相匹配，设定位置变量
    for i in range(row):
        for j in range(col):
            variable_data[j, i] = input_values[k]
            tip = f'第 {j} 列，第 {i} 行,第 {k} 个网格值'
            print(tip)
            k += 1
    with nc.Dataset(output_path, 'a') as ncfile:
        ncfile.variables[tempalate_dataset.variables[template_variable_index].name][:] = variable_data

def get_extent_data(nc_file,variable_name,x_name='lon',y_name='lat', extent=[]):
    '''
    :param nc_file: netcdf 文件路径
    :param variable_name: netcdf文件中要读取变量名称
    :param x_name: netcdf文件中x坐标方向的维度变量名称
    :param y_name: netcdf文件中y坐标方向的维度变量名称
    :param extent: 要提取的数据范围，默认为空，表示提取全域所有数据，如指定范围则按min_longitude，max_longitude，min_latitude，max_latitude顺序进行指定；
        如
        Guangdong Domain 的extent=[109.25,117.75,19,27]；
        China Domain的 extent = [70.25, 136.25, 4.75, 55.25];
        注意：截取后 data_subset.lon 和 data_subset.lat 是一维数组，data_subset 是对应的二维浓度数据
    :return: 返回提取后的数据
    '''

    # 打开NetCDF文件
    ds = xr.open_dataset(nc_file, decode_times=False)
    data = ds[variable_name]
    # 获取变量的单位信息
    unit = data.attrs.get('units')
    # Define the latitude and longitude ranges you want to display
    # min_longitude = ds[x_name].data.min()
    # max_longitude = ds[x_name].data.max()
    # min_latitude = ds[y_name].data.min()
    # max_latitude = ds[y_name].data.max()
    #考虑到有些数据经纬度并不一定是按数值大小进行排序，比如维度它是从90到-90，此时如果算最大值最小值的范围是-90,90，这个范围就会取不到任何数据。所以用维度的第一个数据和最后一个数值表示最小值和最大值
   
    min_longitude = ds[x_name].data[0]
    max_longitude = ds[x_name].data[-1]    
    min_latitude = ds[y_name].data[0]
    max_latitude = ds[y_name].data[-1]

    if extent:  # 不为空则用用户指定的范围       
        min_longitude = extent[0] if min_longitude<max_longitude else extent[1]
        max_longitude = extent[1] if min_longitude<max_longitude else extent[0]
        min_latitude = extent[2] if min_latitude<max_latitude else extent[3]
        max_latitude = extent[3] if min_latitude<max_latitude else extent[2]
        # min_longitude = extent[0] if(min_longitude<max_longitude and ds[x_name].data[0]<ds[x_name].data[1]) else extent[1]
        # max_longitude = extent[1] if min_longitude<max_longitude and ds[x_name].data[0]<ds[x_name].data[1] else extent[0]
        # min_latitude = extent[2] if min_latitude<max_latitude and ds[y_name].data[0]<ds[y_name].data[1] else extent[3]
        # max_latitude = extent[3] if min_latitude<max_latitude and ds[y_name].data[0]<ds[y_name].data[1] else extent[2]
    # 定义要显示的经纬度范围slice(min_longitude, max_longitude)和slice(min_latitude, max_latitude)
    # Use .sel() to select the specific latitude and longitude range
    # 根据维度名称获取维度变量    
    #data_subset = data.sel(x=slice(min_longitude, max_longitude), y=slice(min_latitude, max_latitude))
    data_subset = data.sel({x_name: slice(min_longitude, max_longitude), y_name: slice(min_latitude, max_latitude)})

    return data_subset,unit

def update_netcdf_file(nc_file):
    data = nc.Dataset(nc_file, mode='r+', format="NETCDF4")
    data.variables['PM25_TOT'].units = "μg/m3"    
    # data.TSTEP = 240000
    # data.SDATE  =  data.SDATE+1
    data.close()


def generate_date_array(start_date_str, num_days):
    from datetime import datetime, timedelta
    date_format = "%Y-%m-%d %H:%M:%S"
    start_date = datetime.strptime(start_date_str, date_format)
    date_array = [start_date + timedelta(days=i) for i in range(num_days)]
    date_strings = [date.strftime(date_format) for date in date_array]
    return date_strings

def update_netcdf_file_variable(nc_file,var_name,var_value,var_unit=''):
    data = nc.Dataset(nc_file, mode='r+', format="NETCDF3_CLASSIC")   
    data.variables[var_name][:] = var_value
    if var_unit:
        data.variables[var_name].units =var_unit
    data.close()

def create_netcdf_file(output_file, data, var_name, var_unit):
 
    # 创建NetCDF文件
    ncfile = nc.Dataset('example.nc', 'w')

    # 创建维度
    time_dim = ncfile.createDimension('time', None)
    lat_dim = ncfile.createDimension('lat', 10)
    lon_dim = ncfile.createDimension('lon', 20)

    # 创建变量
    time_var = ncfile.createVariable('time', 'f4', ('time',))
    lat_var = ncfile.createVariable('lat', 'f4', ('lat',))
    lon_var = ncfile.createVariable('lon', 'f4', ('lon',))
    data_var = ncfile.createVariable('data', 'f4', ('time', 'lat', 'lon'))

    # 设置变量的属性
    time_var.units = 'days since 2000-01-01'
    lat_var.units = 'degrees_north'
    lon_var.units = 'degrees_east'
    data_var.units = 'kg/m^2'

    # 设置变量的值
    time_var[:] = [0, 1, 2, 3, 4]
    lat_var[:] = range(10)
    lon_var[:] = range(20)
    data_var[:] = 1.0

    # 关闭NetCDF文件
    ncfile.close()

def create_netcdf_file_with_dims_update(files):    
    # 打开第一个文件以获取变量和维度信息
    with nc.Dataset(files[0], 'r') as src_dataset:      
        file_format=src_dataset.file_format  
        import os
        base_name=os.path.basename(files[0])   
        dir_name=os.path.dirname(files[0]) 
        date_time = datetime.strptime(os.path.basename(dir_name), "%Y%m%d%H")
        file_name = f"{date_time.strftime('%Y%m%d')}.{base_name}.nc"          
        new_file =os.path.join(os.path.dirname(dir_name),file_name)
        # 创建新的合并文件
        with nc.Dataset(new_file, 'w',format=file_format) as dst_dataset:
            # 复制全局属性
            dst_dataset.setncatts(src_dataset.__dict__)
            # 复制维度
            for name, dimension in src_dataset.dimensions.items():
                dst_dataset.createDimension(name, len(dimension) if not dimension.isunlimited() else None)
            # 复制变量
            for name, variable in src_dataset.variables.items():
                # 创建变量
                dst_variable = dst_dataset.createVariable(name, variable.dtype, variable.dimensions)
                # 复制变量属性
                dst_variable.setncatts(variable.__dict__)
                # 创建合并数据数组
                merged_data = np.zeros(variable.shape)   
                # 合并数据
                for file in files:
                    with nc.Dataset(file, 'r') as src_new_dataset:
                        src_variable = src_new_dataset.variables[name]
                        merged_data += src_variable[:]   
                # # 计算每天的平均值             
                daily_mean_data=merged_data/len(files)
                # 将数据写入新的合并文件
                dst_variable[:] = daily_mean_data
            print(f"成功创建{file_name}")    
    
    
if __name__=="__main__":
    # nc_file=r'E:\CO2\data\emis\posterior emission\GD_9km\CONC_BG\2022010100\CONC.bg'
    # model_dir=r'E:\CO2\data\emis\posterior emission\GD_9km\CONC_BG'
    # dates=pd.date_range('2022-01-01',end='2022-12-31',freq='D')
    # from esil.file_helper import get_model_files_by_date
    # for date in dates:
    #     date_str=date.strftime('%Y%m%d')
    #     files=get_model_files_by_date(model_dir,date_str,date_str)
    #     create_netcdf_file_with_dims_update(files)        
      
    
    nc_file='E:/data/emis/CarbonMonitor_total_y2019_m06.nc'
    # start_date = "2019-06-01 00:00:00"
    # num_days = 30
    # date_array =generate_date_array(start_date, num_days)
    # date_array = np.linspace(0, 29, 30)
    # var_unit= f"days since {start_date}"   
    var_name='nday' 
    # update_netcdf_file_variable(nc_file,var_name,date_array,var_unit)
    # data = nc.Dataset(nc_file, mode='r', format="NETCDF3_CLASSIC")
    from esil.file_helper import get_files
     # 获取目录中的所有文件
    variable_name='emission'
    files = get_files(r'E:\data\emis\GRACED')
    all_data = []
    # 逐个打开每个文件并读取数据
    for file in files:
        dataset = nc.Dataset(file)
        variable_data = dataset.variables[variable_name][:]  # 替换为您要读取的变量名称
        all_data.append(variable_data)

    # 合并所有数据为一个大数组
    all_data = np.concatenate(all_data, axis=0)
    # 计算全年平均
    annual_mean = np.mean(all_data, axis=0)
    # 将结果保存到新的文件中
    new_file = "annual_mean.nc"
    new_dataset = nc.Dataset(new_file, mode="w", format="NETCDF4")
    new_dataset.createDimension("time", annual_mean.shape[0])   
    new_dataset.createVariable("annual_mean", "f4", ("time",))
    new_dataset.variables["annual_mean"] = annual_mean
    new_dataset.close()
    



    # 指定要合并的维度 
    # 使用 MFDataset 函数打开多个文件并指定要合并的维度
    # dataset =nc.MFDataset(files, aggdim=var_name)
    # f = nc.MFDataset(files)
    # var = f.variables[variable_name][:]

    nc_file=r'E:\new_CT2022.molefrac_components_glb3x2_2020-12-31.nc'
     
    from plot_map import plotmap
    nc_file= r'/NetcdfReader/CO2 Visualization/CO2-em-anthro_input4MIPs_emissions_CMIP_CEDS-2021-04-21_gn_200001-201912.nc'
    variable_name='CO2_em_anthro'
    data,unit = get_extent_data(nc_file, variable_name)
    #data=get_extent_data(nc_file,variable_name)
    # 对指定列进行求和，使用 axis 参数来指定列维度
    #sum_sector_data = np.sum(data, axis=1)
    sum_data=np.sum(data,axis=(0,1))
    x = sum_data.lon.values
    y = sum_data.lat.values
    data = sum_data.values
    grid_x, grid_y = np.meshgrid(x, y)
    fig=plotmap(grid_x, grid_y, grid_concentration=data, title=variable_name)
    if fig:
        fig.show()




import sympy as sp

def O3_function(NOx,VOCs,SO2,NH3,k_values):
    '''
    :param NOx: 数值=Emission ratio-1
    :param VOCs: 数值=Emission ratio-1
    :param SO2: 数值=Emission ratio-1
    :param NH3: 数值=Emission ratio-1
    :param k_values: 顺序：[NOx^5	NOx^4	NOx^3	NOx^2	NOx	VOC	VOC^2	VOC^3	NOx*VOC	NOx*VOC^3	NOx^5*VOC	NOx^2*VOC	SO2	NH3	baseCase]
    :return:
    '''
    return   k_values[0] * NOx ** 5 + k_values[1] * NOx ** 4 + k_values[2] * NOx ** 3 + k_values[3] * NOx ** 2 + k_values[4] * NOx + k_values[5] * VOCs + k_values[6] * VOCs ** 2 + k_values[7] * VOCs ** 3 \
             + k_values[8] * NOx * VOCs + k_values[9] * NOx * VOCs ** 3+ k_values[10] * NOx ** 5 * VOCs + k_values[11] * NOx ** 2 * VOCs  + k_values[12] * SO2 + k_values[13] * NH3+ k_values[14]

def get_O3_PR_value(k_values,VOCs=1,SO2=1,NH3=1):
    '''
    :param k_values: 顺序：[NOx^5	NOx^4	NOx^3	NOx^2	NOx	VOC	VOC^2	VOC^3	NOx*VOC	NOx*VOC^3	NOx^5*VOC	NOx^2*VOC	SO2	NH3	baseCase]
    :param VOCs: Emission ratio
    :param SO2: Emission ratio
    :param NH3: Emission ratio
    :return:
    '''
    # 解方程并找到使得 O3 达到最大值的 NOx 值
    VOCs_symbol, NOx_symbol,SO2_symbol,NH3_symbol = sp.symbols('VOCs NOx SO2 NH3')
    O3_expression = O3_function(NOx_symbol,VOCs_symbol, SO2_symbol,NH3_symbol,k_values)
    dO3_dNOx = sp.diff(O3_expression, NOx_symbol)
    max_NOx_solution = sp.solve(dO3_dNOx, NOx_symbol)
    max_NOx_value=max_NOx_solution[0]
    max_NOx_value=max_NOx_value.subs(VOCs_symbol, VOCs - 1)

    # 要进行替换的多个变量和值
    substitutions = {NOx_symbol: max_NOx_value, SO2_symbol: SO2-1,NH3_symbol:NH3-1,VOCs_symbol:VOCs-1}  # 替换为 max_NOx_value 和 5
    # 对表达式进行多个变量的替换
    max_O3_result = O3_expression.subs(substitutions)
    NOx_emission_ratio=max_NOx_value+1# max_NOx_value=NOx_emission_ratio-1,其中1表示基准，所以换算回NOx_emission_ratio，需要将max_NOx_value+1.
    return NOx_emission_ratio, max_O3_result

def get_O3_VNr(k_values,VOCs,NOx):
    '''
    :param k_values:  顺序：[NOx^5	NOx^4	NOx^3	NOx^2	NOx	VOC	VOC^2	VOC^3	NOx*VOC	NOx*VOC^3	NOx^5*VOC	NOx^2*VOC	SO2	NH3	baseCase]
    :param VOCs: 数值=Emission ratio-1
    :param NOx: 数值=Emission ratio-1
    :return:
    '''
    expression=-(5*k_values[0]*NOx**4+4*k_values[1]*NOx**3+3*k_values[2]*NOx**2+2*k_values[3]*NOx+k_values[4]+5*k_values[10]*NOx**4*VOCs+2*k_values[11]*NOx*VOCs+k_values[8]*VOCs+k_values[9]*VOCs**3)/(k_values[10]*NOx**5+k_values[11]*NOx**2+k_values[8]*NOx+3*k_values[9]*NOx*VOCs**2+k_values[5]+2*k_values[6]*VOCs+3*k_values[7]*VOCs**2)
    return expression

if __name__=="__main__":

    # 示例参数
    import pandas as pd
    df=pd.read_csv(r'/NetcdfReader/O3_Monitor_Functions.csv')
    # 将第一列之后的所有列设定为 float 类型
    for col in df.columns[1:]:
        df[col] = df[col].astype(float)
    # 遍历每一行数据
    for index, row in df.iterrows():
        # 使用 iloc 来获取除第一列外的其他列的值，并转换为数组
        row_values = row.iloc[1:].values
        VOCs_value=1,
        #row_values = [-0.1018906, 0.06396596, 0.03682716, -0.1217417, 0.06571316, 0.3468532, -0.2358255, -0.08597394,                    0.6097692, -0.04344897, -0.06851652, 0.115323, -0.03890517, -0.08019463,99]
        result_NOx, result_O3 = get_O3_PR_value(row_values.tolist())

        #vnr=get_O3_VNr(row_values,-0.5,-0.5)
        #print(f"网格{row[0]},VNr(0.5,0.5)={vnr},NOx={result_NOx},O3浓度值最大({result_O3})")
        print(f"网格{row[0]},VNr(0,0)={-row_values[4]/row_values[5]},O3浓度值最大({result_O3})")


    # k_values=np.random.rand(14)
    # k_values = [0.1] * 14
    # k_values=[-0.1018906,0.06396596,0.03682716,-0.1217417,0.06571316,0.3468532,-0.2358255,-0.08597394,0.6097692,-0.04344897,-0.06851652,0.115323,-0.03890517,-0.08019463,99]
    # VOCs_value = 2.0  # 示例 VOCs 的数值
    # result_NOx, result_O3 = get_O3_PR_value(k_values)
    # SO2_symbol, NH3_symbol = sp.symbols('SO2 NH3')
    # result_O3 = result_O3.subs(SO2_symbol, 2)
    # O3_2 = result_O3.subs(NH3_symbol, 2)
    #

    # k1 = 0.1
    # k2 = 0.2
    # k3 = 0.15
    # # 将得到的最终表达式中的 VOCs 变量代入数值并计算结果
    # #O3_expression_specifiedVOCs = O3_expression.subs(VOCs_symbol, VOCs_value)
    #
    # # 调用函数并传入参数
    # result_NOx, result_O3 = get_differentiate(k1, k2, k3, VOCs_value)
    # print(f"使得 O3 达到最大值的 NOx 值为: {result_NOx}")
    # print(f"最终表达式结果为: {result_O3}")
    #
    #
    # a=get_differentiate(1)
    # # 定义符号变量
    # VOCs, NOx = sp.symbols('VOCs NOx')
    #
    #
    # # 计算当 VOCs = 1 时，使得 O3 数值最大的 NOx 数值
    # VOCs_value = 1
    # #
    # print(f'当 VOCs = {VOCs_value} 时，使得 O3 数值最大的 NOx 数值为 {max_O3_NOx}，对应的最大 O3 数值为 {max_O3_value}')



'''
Author: Devin
Date: 2024-02-27 22:50:27
LastEditors: Devin
LastEditTime: 2024-04-14 11:02:47
FilePath: \PythonDemo\esil\file_helper.py
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
import os

def get_files(root_directory, file_format='',recursive=True):
    '''
    @description: 获取指定文件夹下的所有文件路径
    @param {str} root_directory，需要获取的文件夹目录
    @param {str} file_format，需要获取的文件格式，如'.txt'
    @param {bool} recursive，是否递归遍历子文件夹
    @param {bool} is_include_dir，是否包含文件夹路径
    @return {list}
    @eg: get_files(root_directory='./', file_format='.txt', recursive=True, is_include_dir=False)
    @author: Devin
    @date: 2022-03-01 16:30:23
    '''
    all_files = []
    for root, dirs, files in os.walk(root_directory):       
        for file in files:
            if file_format == '':
                all_files.append(os.path.join(root, file))
            elif file.endswith(file_format):                
                all_files.append(os.path.join(root, file))    
        if not recursive:
            dirs.clear()  # 清空子目录列表，以便不进入子目录
    return all_files

def get_files_old(dir):
    '''
    @description: 获取指定文件夹下的所有文件路径;已废弃
    @param {str} dir，需要获取的文件夹目录
    @return {list}
    @eg: get_files_old(dir='./')   
    '''
    files = os.listdir(dir)
    file_list = []
    for file in files:
        file_path = os.path.join(dir, file)
        # 判断是否是文件夹
        if os.path.isdir(file_path):
            continue
        else:
            file_list.append(file_path)           
    return file_list


def get_files_by_format(directory, file_format='',is_include_dir=False):
    all_files = []
    import glob
    # # 递归遍历目录及其子目录
    # for file_path in glob.iglob(directory + '/**/*.' + file_format, recursive=True):
    #     if is_include_dir:
    #         all_files.append(file_path)
    #     else:
    #         if not os.path.isdir(file_path):
    #             # 排除目录，只添加文件路径
    #             all_files.append(file_path)        
    # 递归遍历目录及其子目录
    for root, dirs, files in os.walk(directory):
        for file in files: 
            if file.endswith(file_format):
                file_path = os.path.join(root, file)
                all_files.append(file_path)
    return all_files


def get_model_files_by_date(model_dir,start_date,end_date):
    '''
    @description: 根据日期获取指定日期范围内的模型文件
    @param {str} model_dir，模型文件夹路径
    @param {str} start_date，开始日期，如'20220301'
    @param {str} end_date，结束日期，如'20220301'
    @return {list} 模型文件路径列表
    @eg: get_model_files_by_date(model_path,'20220301','20220301')
    @author: Devin
    @date: 2022-03-01 16:30:23
    '''       
    from datetime import datetime,timedelta
    start_date = datetime.strptime(start_date, '%Y%m%d')
    end_date = datetime.strptime(end_date, '%Y%m%d')
    end_date=end_date+timedelta(hours=23)
    date_format = "%Y%m%d%H"
    res_files=[]
    for root, dirs, files in os.walk(model_dir):       
        for directory in dirs:
            date_time = datetime.strptime(directory, date_format)
            if date_time >= start_date and date_time <= end_date:
                sub_dir = os.path.join(root, directory)
                sub_files=get_files(sub_dir,recursive=False)
                res_files.append(sub_files[0])  
    return res_files 




'''
Author: Devin
Date: 2024-05-01 12:13:42
LastEditors: Devin
LastEditTime: 2024-05-01 12:36:32
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
import json
import numpy as np
import xarray as xr
from shapely.geometry import shape, Point, MultiPolygon
from shapely.prepared import prep
from tqdm.auto import tqdm#显示进度
import os
import pickle
#解释aox函数    
#func是一个函数，n是一个整数
#返回一个新的函数，这个函数的作用是对输入的数组进行转置，然后对转置后的数组进行func操作 
#再对func操作后的数组进行转置
#perm是一个数组，这个数组的作用是对数组进行转置
def aox(func, n=0):
    def wrapper(x):
        perm = [n] + [0 if i==n else i for i in range(1,x.ndim)]
        tr = lambda arr: np.transpose(arr, perm)
        return tr(func(tr(x)))
    return wrapper
#计算网格点的均值
mean = lambda arr: (arr[1:,...] + arr[:-1,...]) / 2
#计算网格点的面积
ext = lambda arr: np.concatenate([[2*arr[0,...] - arr[1,...]], arr, 2*arr[-1,...] - [arr[-2,...]]  ], axis=0)
dup = lambda func: (lambda x: aox(func, 0)(aox(func, 1)(x)) )
#计算网格点的面积
stag = dup(lambda arr: mean(ext(arr)))
#将角度转化为弧度
deg2rad = lambda deg: deg/180*np.pi

#计算经纬度网格点的梯度
#axis=0时，计算纬度方向的梯度
#axis=1时，计算经度方向的梯度
def get_del_vect(lon, lat, axis=0):
    '''
    @description: 计算经纬度网格点的梯度
    @param (np.array) lon: 经度坐标
    @param (np.array) lat: 纬度坐标 
    @param (int) axis: 计算纬度方向的梯度时，axis=0；计算经度方向的梯度时，axis=1
    @return {np.array} dx, dy: 经度方向的梯度和纬度方向的梯度
    '''
    a = 6371e3
    d = aox((lambda arr: arr[1:,...] - arr[:-1,...]), axis)
    h = aox((lambda arr: arr[:-1,...]), 1-axis)
    t = lambda arr: deg2rad(h(d(stag(arr))))
    dx = a * t(lon) * np.cos(deg2rad(lat))
    dy = a * t(lat)
    return dx, dy

def get_m2(lon, lat):
    '''
    @description: 计算网格点的面积
    @param (np.array) lon: 经度坐标
    @param (np.array) lat: 纬度坐标
    @return {np.array} 面积（单位m2）
    '''
    #计算网格点的梯度
    v0_dx, v0_dy = get_del_vect(lon, lat, axis=0)
    #将网格点的梯度转置
    v1_dx, v1_dy = get_del_vect(lon, lat, axis=1)
    #计算网格点的面积,v1_dx*v0_dy - v1_dy*v0_dx 为什么是相减？
    return np.abs( v1_dx*v0_dy - v1_dy*v0_dx )

def get_mask(json_file, lon, lat,replaceFalseWithNan=False):
    '''
    @description: 从geojson文件中获取mask
    @param (str) json_file: geojson文件路径
    @param (np.array) lon: 经度坐标
    @param (np.array) lat: 纬度坐标
    @return {np.array} mask: mask
    '''
    with open(json_file, 'r', encoding='utf-8') as f:
        poly = shape(json.load(f)['features'][0]['geometry'])
    poly = sorted(poly.geoms, key=lambda p: p.area, reverse=True)
    #path = Path.make_compound_path(*geos_to_path(poly[:3]))   
    poly = prep(MultiPolygon(poly[:3]))
    mask = list(map(lambda x, y: poly.contains(Point(x, y)), tqdm(lon.flat), lat.flat))
    mask = np.array(mask).reshape(lon.shape)
    if replaceFalseWithNan:       
        # 将False替换为NAN
        mask = np.where(mask == False, np.nan, mask) 
    return mask

def get_all_masks(json_file, lon, lat,replaceFalseWithNan=False):
    '''
    @description: 从geojson文件中获取所有mask
    @param (str) json_file: geojson文件路径
    @param (np.array) lon: 经度坐标
    @param (np.array) lat: 纬度坐标
    @return {dict_masks: dict}
    '''
    dict_masks={}
    with open(json_file, 'r', encoding='utf-8') as f:
        for i,feature in enumerate(json.load(f)['features']):
            poly = shape(feature['geometry'])
            poly = sorted(poly.geoms, key=lambda p: p.area, reverse=True)
            poly = prep(MultiPolygon(poly[:3]))
            mask = list(map(lambda x, y: poly.contains(Point(x, y)), tqdm(lon.flat), lat.flat))
            mask = np.array(mask).reshape(lon.shape)
            if replaceFalseWithNan:       
                # 将False替换为NAN
                mask = np.where(mask == False, np.nan, mask)
            name=feature['properties']['name'] if 'name' in feature['properties'] else str(i)
            dict_masks[name]=mask        
    return dict_masks

def get_mask_with_name(json_file, lon, lat,field_name='adcode',replaceFalseWithNan=False): 
    '''
    @description: 从geojson文件中获取mask，并返回mask的名称
    @param (str) json_file: geojson文件路径
    @param (np.array) lon: 经度坐标
    @param (np.array) lat: 纬度坐标
    @return {np.array} mask_name: mask的名称
    @return {np.array} mask: mask    
    @return {dict} dict_data: 包含每个mask的名称的字典
    '''   
    # 创建一个空的 mask，初始化为 -1，表示没有匹配的 feature
    mask = np.full(lon.shape, -1, dtype=int)  
    mask_name = np.full(lon.shape, "", dtype='U10')#避免字符串被截取掉，需要设定一个字符串长度10，而不能用默认的str类型
    #mask_name = np.full(lon.shape, "", dtype='str')
    dict_data={}  
    with open(json_file, 'r', encoding='utf-8') as f:
        for idx,feature in tqdm(enumerate(json.load(f)['features'])):
            poly = shape(feature['geometry'])
            poly = sorted(poly.geoms, key=lambda p: p.area, reverse=True)
            poly = prep(MultiPolygon(poly[:3]))
            # mask = list(map(lambda x, y: poly.contains(Point(x, y)), tqdm(lon.flat), lat.flat))
            # mask = np.array(mask).reshape(lon.shape)            
            contains_mask = np.array(list(map(lambda x, y: poly.contains(Point(x, y)), lon.flat, lat.flat)))
            contains_mask = contains_mask.reshape(lon.shape)
            mask[contains_mask] = idx  # 将属于当前多边形的点的 mask 设置为该多边形的索引            
            # if replaceFalseWithNan:       
            #     # 将False替换为NAN
            #     mask = np.where(mask == False, np.nan, mask) 
            if field_name in feature['properties']:
                name=feature['properties'][field_name]
            elif 'adcode' in feature['properties']:
                name=feature['properties']['adcode']
            else:
                name=str(idx)            
            dict_data[idx]=name
            mask_name[contains_mask] = name   
    if replaceFalseWithNan:
        # 将 -1 替换为 NaN
        mask = np.where(mask == -1, np.nan, mask)         
    return mask_name,mask,dict_data

def get_boundary_mask_data(boundary_json_file,lats, lons,field_name='adcode',replaceFalseWithNan=True,pkl_file_name=None):  
    '''
    @description: 从geojson文件中获取边界mask数据
    @param (str) boundary_json_file: 边界geojson文件路径
    @param (np.array) lats: 纬度坐标
    @param (np.array) lons: 经度坐标
    @param (bool) replaceFalseWithNan: 是否将False替换为NAN
    @return {dict_masks: dict}
    @return {np.array} mask_name: mask的名称
    '''  
    dir=os.path.dirname(boundary_json_file)
    boundary_file_name=os.path.basename(boundary_json_file)
    boundary_mask_file=os.path.join(dir,f'{boundary_file_name}.pkl') if pkl_file_name is None else os.path.join(dir,pkl_file_name)
    if os.path.exists(boundary_mask_file):
        with open(boundary_mask_file, 'rb') as file:
            dict_masks,mask_name = pickle.load(file)
    else:        
        dict_masks=get_all_masks(json_file=boundary_json_file,lat=lats,lon=lons,replaceFalseWithNan=replaceFalseWithNan)
        mask_name,_,_=get_mask_with_name(json_file=boundary_json_file,lat=lats,lon=lons,field_name=field_name,replaceFalseWithNan=replaceFalseWithNan) 
        masks=(dict_masks,mask_name)
        # 将字典保存到文件
        with open(boundary_mask_file, 'wb') as file:
            pickle.dump(masks, file)
    return dict_masks,mask_name

if __name__ == '__main__':
    from wrf import (to_np, getvar, smooth2d, get_cartopy, cartopy_xlim,
                 cartopy_ylim, latlon_coords,get_basemap)
    import netCDF4 as nc 
    
    
    wrf_out_file=r'/work/home/pengzhen/PRD9km_2/devin/python_project/code/co2_inversion/data/wrfinput_d01.nc'
    dataset = nc.Dataset(wrf_out_file, 'r') 
    t2 = getvar(dataset, "T2")
    # Get the cartopy mapping object
    cart_proj = get_cartopy(t2)
    # Get the latitude and longitude points
    lats, lons =latlon_coords(t2)# latlon_coords(t2)  
    lons,lats =to_np(lons), to_np(lats)  
    boundary_json_file='/work/home/pengzhen/PRD9km_2/devin/python_project/code/co2_inversion/data/boundary/guangdong_cities.json'
    dict_mask,masknames=get_boundary_mask_data(boundary_json_file,lats, lons)   
    print('done')
    # print(dict_masks1)
    # print(mask_name1)
    # 显示多个地图


import pandas as pd
from sqlalchemy import create_engine, text
import requests
from datetime import datetime, timedelta
import os
import time
# import logging
from logging.handlers import TimedRotatingFileHandler
import sqlite3
import numpy as np
from log_helper import get_logger
import random
import string
from tqdm import tqdm
# 获取当前脚本所在的目录
script_dir = os.path.dirname(os.path.abspath(__file__))
# 设置新的工作目录
os.chdir(script_dir)

# 创建一个logger
log_name = f'./logs/sqlite_helper.log'
logger = get_logger(log_name)


def validate_data_columns(str_sqlite_db,table_name, data_columns):
    ''' 
    @description: 检查数据列数是否与表的列数匹配
    @param {str} str_sqlite_db: SQLite数据库文件路径
    @param {str} table_name: 表名
    @param {list} data_columns: 数据列名列表
    @return {bool} 成功返回True，失败返回False
    '''
    conn = sqlite3.connect(str_sqlite_db)
    cursor = conn.cursor()
    # 获取表的列数
    cursor.execute(f"PRAGMA table_info({table_name})")
    num_columns = len(cursor.fetchall())
    # 检查数据的列数是否与表的列数匹配
    if len(data_columns) != num_columns:
        logger.error(f"Error: Number of columns in data ({len(data_columns)}) does not match the number of columns in table ({num_columns})")
        return False
    else:
        return True
    
def save_df_to_db(str_sqlite_db,df,table_name,if_exists='replace',index=False,identified_columns=None,use_temp_table=True,batch_size=10000):  #='station_info'
    '''
    @description: 将数据框保存到数据库
    @param {str} str_sqlite_db: SQLite数据库文件路径
    @param {pd.DataFrame} df: 数据框
    @param {str} table_name: 表名
    @param {str} if_exists:replace; append; fail; default='replace' 若存在则覆盖，否则新建;
    @param {str} identified_columns: 索引标签, default=None,若指定数据库表中的主键列名称。通过指定主键列，系统将根据主键列的值来判断是否插入新数据；尚不存在的新记录时，才会将这些新记录插入到数据库表中。已存在的相同记录将不会被重复插入
    @return {bool} 成功返回True，失败返回False
    '''
    try:
        conn = sqlite3.connect(f"{str_sqlite_db}") 
        cursor = conn.cursor()           
        cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}' COLLATE NOCASE")#COLLATE NOCASE 忽略大小写
        result = cursor.fetchone()
        table_exists = result is not None
        if table_exists:
            is_ok=validate_data_columns(str_sqlite_db,table_name, df.columns.tolist())
            if not is_ok:  
                # 列数不匹配时，重命名当前表
                new_table_name = f'{table_name}_{datetime.now().strftime("%Y%m%d%H%M%S")}'
                cursor.execute(f"ALTER TABLE {table_name} RENAME TO {new_table_name}")
                # 创建新表来存储数据
                df.to_sql(table_name, conn, if_exists=if_exists, index=index)
                conn.close()
                return True
        if identified_columns is not None:    
            if not table_exists:
                # 表不存在，创建表
                df.to_sql(table_name, conn, if_exists=if_exists, index=index)
                conn.close()
                return True
            if not use_temp_table:
                #  # 检查数据是否已存在于表中
                # 构建需要选择的字段            
                select_columns = ",".join(identified_columns)
                existing_data_query = f"SELECT {select_columns} FROM {table_name} WHERE " + " AND ".join([f"{col} = ?" for col in identified_columns])
                not_existing_data = []
                for index, row in df.iterrows():                
                    cursor.execute(existing_data_query,  [row[col] for col in identified_columns])
                    result = cursor.fetchall()
                    if len(result) == 0:
                        not_existing_data.append(row)            
                # 插入不存在的数据
                if not_existing_data:
                    new_data_df = pd.DataFrame(not_existing_data)
                    new_data_df.to_sql(f'{table_name}', conn, if_exists=if_exists, index=index)
                    logger.info("插入数据成功！")
                else:
                    logger.info("所有数据已存在，无需插入。")
            else:# 对于数据量较大的表，使用临时表：将要插入的数据存储在临时表中，然后执行一次查询来过滤出已存在的数据，最后将剩余的数据插入到目标表中。                
                # 按批次插入数据
                for i in tqdm(range(0, len(df), batch_size),leave=False,desc= "分批插入数据"):
                    batch_data_df = df[i:i+batch_size]
                    insert_unique_data(conn, batch_data_df, table_name, identified_columns)
                if len(df) > 100000:#数据条数超过十万，需要 VACUUM 命令来重建数据库文件，并清除未使用的空间，从而减小数据库文件的大小
                    logger.info(f"start to VACUUM {str_sqlite_db}")#cursor.rowcount获取生效的条数  
                    # 执行 VACUUM 命令
                    cursor.execute("VACUUM")
                    # 提交更改
                    conn.commit()
            conn.close()
        else:
            # 建立与SQLite数据库的连接
            engine = create_engine(f'sqlite:///{str_sqlite_db}')  # 这里的"data.db"是SQLite数据库文件的路径
            # 将DataFrame数据插入数据库表
            df.to_sql(table_name, con = engine.connect(), if_exists=if_exists, index=index)   
            # 关闭数据库连接
            engine.dispose()
        logger.info(f"save {table_name} to db done") 
        return True 
    except Exception as e:
            logger.error(f"save {table_name} to db failed", e)     
            return False  
        
def rebulid_db(str_sqlite_db):
    '''
    @description: 重建数据库
    @param {str} str_sqlite_db: SQLite数据库文件路径
    '''
    conn = sqlite3.connect(f"{str_sqlite_db}") 
    cursor = conn.cursor()
    logger.info(f"start to VACUUM {str_sqlite_db}") 
    # 执行 VACUUM 命令
    cursor.execute("VACUUM")
    # 提交更改
    conn.commit()
    conn.close()           
                  
def insert_unique_data(conn, df, table_name, unique_columns):
    '''
    @description: 插入唯一数据:使用临时表：将要插入的数据存储在临时表中，然后执行一次查询来过滤出已存在的数据，最后将剩余的数据插入到目标表中。
    @param {sqlite3.Connection} conn: 数据库连接
    @param {pd.DataFrame} df: 数据框
    @param {str} table_name: 表名
    @param {list} unique_columns: 唯一列名列表
    '''
    try:
        letters = string.ascii_letters
        random_name=''.join(random.choice(letters) for _ in range(8)) 
        # 创建临时表    
        temp_table_name = f'{table_name}_{random_name}'        
        # 创建临时表并将数据插入临时表
        df.to_sql(temp_table_name, conn, if_exists='replace', index=False)
        # 构建唯一记录字段的查询条件
        unique_conditions = " AND ".join([f"{table_name}.{col} = {temp_table_name}.{col}" for col in unique_columns])
        # 使用一次查询来检查并插入数据
        # query = f"""
        # INSERT INTO {table_name}
        # SELECT *
        # FROM {temp_table_name}
        # WHERE NOT EXISTS (
        #     SELECT 1
        #     FROM {table_name}
        #     WHERE {unique_conditions}
        # )
        # """      
        # 改用left join提高sql执行效率
        condition_field = f't2.{unique_columns[0]} IS NULL'
        join_conditions = ' AND '.join([f't1.{col} = t2.{col}' for col in unique_columns])
        query = f''' 
                    INSERT INTO {table_name}                   
                    SELECT t1.* FROM {temp_table_name} t1
                    LEFT JOIN {table_name} t2 ON {join_conditions}
                    WHERE {condition_field}
                    '''
        cursor = conn.cursor()
        cursor.execute(query)
        conn.commit()  
        logger.info(f"insert {cursor.rowcount} records to {table_name} done")#cursor.rowcount获取生效的条数        
        # 删除临时表
        conn.execute(f"DROP TABLE {temp_table_name}")
        # if len(df) > 10000:#数据条数超过一万，需要 VACUUM 命令来重建数据库文件，并清除未使用的空间，从而减小数据库文件的大小
        #     logger.info(f"start to VACUUM {str_sqlite_db}")#cursor.rowcount获取生效的条数  
        #     # 执行 VACUUM 命令
        #     cursor.execute("VACUUM")
        #     # 提交更改
        #     conn.commit()
        return True
    except Exception as e:
            logger.error(f"insert unique data to {table_name} failed", e)     
            return False  
    
def append_df_to_db(str_sqlite_db,df,table_name):
    '''
    @description: 将数据框追加到数据库
    @param {str} str_sqlite_db: SQLite数据库文件路径
    @param {pd.DataFrame} df: 数据框
    @param {str} table_name: 表名
    @return {bool} 成功返回True，失败返回False
    '''
    try:  
        # 连接到 SQLite 数据库
        conn = sqlite3.connect(str_sqlite_db)    
        # 将 DataFrame 中的数据追加到 SQLite 数据库表中
        df.to_sql(table_name, conn, if_exists='append', index=False)
        conn.commit()
        # 关闭数据库连接
        conn.close() 
        return True
    except Exception as e:
            logger.error(f"append {table_name} to db failed", e)     
            return False    
    
def query_db(str_sqlite_db,sql):
    '''
    @description: 查询数据库
    @param {str} str_sqlite_db: SQLite数据库文件路径
    @param {str} sql: SQL语句
    @return {pd.DataFrame} 查询结果数据框;  查询失败返回空数据框 
    '''
    try:    
        # 建立与SQLite数据库的连接
        engine = create_engine(f'sqlite:///{str_sqlite_db}')  # 这里的"data.db"是SQLite数据库文件的路径
         # 读取数据库表
        con = engine.connect()  # 获取数据库连接对象
        df = pd.read_sql(text(sql), con=con)  # 使用连接对象执行查询操作 #需要用sqlalchemy.text(sql)转为sql对象,否则ObjectNotExecutableError reading from
        # 读取数据库表
        #df = pd.read_sql(sql, engine)    
        # 关闭数据库连接
        engine.dispose()
        return df
    except Exception as e:
        logger.error(f"failed to query {str_sqlite_db} with {sql}", e)     
        return pd.DataFrame()       



'''
Author: Devin
Date: 2024-03-09 11:11:25
LastEditors: Devin
LastEditTime: 2024-04-22 22:21:15
FilePath: \PythonDemo\esil\PaneLayout.py
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
# Description: This is a python script to calculate the number of columns and rows for the pane layout.
#description: Define the auto layout options for the pane layout.
class PaneLayout():
    # Layout the pane with a square grid: The number of panes in the row and column are the same.
    ForceSquare = 1
    # Layout the pane with a square grid: The number of panes in the row and column are the same. If the number of panes is not a perfect square, the number of panes in the row is one less than the number of panes in the column.
    SquareColPreferred = 2
    # Layout the pane with a square grid: The number of panes in the row and column are the same. If the number of panes is not a perfect square, the number of panes in the column is one less than the number of panes in the row.
    SquareRowPreferred = 3
    # Layout the pane with a single row: The number of panes in the row is 1 and the number of panes in the column is the same as the number of panes.
    SingleRow = 4
    # Layout the pane with a single column: The number of panes in the row is the same as the number of panes and the number of panes in the column is 1.
    SingleColumn = 5
    # Layout the pane with an explicit number of columns: The first row has 1 column and the second row has 2 columns for a total of 3 panes.
    ExplicitCol12 = 6
    # Layout the pane with an explicit number of columns: The first row has 2 columns and the second row has 1 column for a total of 3 panes.
    ExplicitCol21 = 7
    # Layout the pane with an explicit number of columns: The first row has 2 columns and the second row has 3 columns for a total of 5 panes.
    ExplicitCol23 = 8
    # Layout the pane with an explicit number of columns: The first row has 3 columns and the second row has 2 columns for a total of 5 panes.
    ExplicitCol32 = 9
    # Layout the pane with an explicit number of rows: The first column has 1 row and the second column has 2 rows for a total of 3 panes.
    ExplicitRow12 = 10
    # Layout the pane with an explicit number of rows: The first column has 2 rows and the second column has 1 row for a total of 3 panes.
    ExplicitRow21 = 11
    # Layout the pane with an explicit number of rows: The first column has 2 rows and the second column has 3 rows for a total of 5 panes.
    ExplicitRow23 = 12
    # Layout the pane with an explicit number of rows: The first column has 3 rows and the second column has 2 rows for a total of 5 panes.
    ExplicitRow32 = 13
    ExplicitRow34 =14

#description: 根据count和panel_layout返回col和row
#param count: int
#param panel_layout: PaneLayout,可以是整数，元组，或者None, tuple, None, default None
#return: col(int), row(int)
def get_layout_col_row(count, panel_layout=None):
    if count != 0:        
        if isinstance(panel_layout, tuple):#如果是元组，直接返回
            col,row=panel_layout
        else:
            num = int((count ** 0.5) + 0.9999999)
            if panel_layout == PaneLayout.ForceSquare:
                row = num
                col = num
            elif panel_layout == PaneLayout.SquareColPreferred:
                row = num
                if count <= num * (num - 1):
                    row -= 1               
            elif panel_layout == PaneLayout.SquareRowPreferred:
                row = num
                col = num
                if count <= num * (num - 1):
                    col -= 1
            elif panel_layout == PaneLayout.SingleRow:
                row = 1
                col = count
            elif panel_layout == PaneLayout.SingleColumn:
                row = count
                col = 1       
            elif panel_layout == PaneLayout.ExplicitCol12:
                col = 1
                row = 2
            elif panel_layout == PaneLayout.ExplicitCol21:   
                col = 2
                row = 1
            elif panel_layout == PaneLayout.ExplicitCol23:
                col = 2
                row = 3
            elif panel_layout == PaneLayout.ExplicitCol32:
                col = 3
                row = 2
            elif panel_layout == PaneLayout.ExplicitRow12:
                row = 1
                col = 2
            elif panel_layout == PaneLayout.ExplicitRow21:
                row = 2
                col = 1
            elif panel_layout == PaneLayout.ExplicitRow23:
                row = 2
                col = 3
            elif panel_layout == PaneLayout.ExplicitRow32:
                row = 3
                col = 2 
            elif panel_layout == PaneLayout.ExplicitRow34:
                row = 3
                col = 4 
            elif panel_layout == None and count in (3,5,7):   
                row = 1
                col = count         
            else:#include none and others
                row = num
                col = num  
                if count <= num * (num - 1):
                    row -= 1  
        return col, row

if __name__ == '__main__':
    count = 8
    pane_layout = PaneLayout.ForceSquare    
    col,row= get_layout_col_row(count, pane_layout)
    print(col, row)
    
    col1,row1= get_layout_col_row(count, None)
    print(col1, row1)
    col2,row2= get_layout_col_row(count, (4,2))
    print(col2, row2)
    col3,row3= get_layout_col_row(count, (3,3))
    print(col3, row3)
    col4,row4= get_layout_col_row(count, 5)
    print(col4, row4)
 


'''
Author: Devin
Date: 2024-06-13 11:49:52
LastEditors: Devin
LastEditTime: 2024-06-18 23:53:53
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
'''
from setuptools import setup, find_packages

setup(
    name='esil',
    version='1.0.1',
    author='Devin Long',
    author_email='long.sc@qq.com',
    description='commonly used functions writted by Devin Long',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/Devin-Long-7/esil',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    # include_package_data=True,# 允许包含在 MANIFEST.in 文件中指定的所有文件。
    # package_data={
    #     'esil': ['*.py'],
    #     'esil.rsm': ['RSM/*.py'],
    #     # 'devin_colors': ['color_maps/*.pkl'],  # 包含特定包中的 .pkl 文件
    # },
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'xarray',
        'scipy',
        'tqdm',
        'netCDF4',
        'pyproj',
        'SQLAlchemy',     
        'chardet',
        'sympy',
    ],
    #     # rich >= 9.13.0
    # install_requires =
    # numpy
    # matplotlib
    # ,
)


import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.io.shapereader import Reader
import cartopy.mpl.geoaxes
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
import geopandas as gpd
import numpy as np
from PaneLayout import get_layout_col_row
import math
#for wrf_out data

# 自定义函数，根据数据的大小选择保留不同位数的小数
def format_data(x):
    if abs(x) > 1000:
        return '{:.0f}'.format(x)  # 大于1000的数据保留0位小数
    elif abs(x) >1 :
        return '{:.2f}'.format(x)  # 大于1的数据保留2位小数
    elif abs(x) < 0.001:
        return '{:.2e}'.format(x)  # 小于0.001的数据保留2位小数，采用科学技术法
    else:
        return '{:.3f}'.format(x)  # 大于0.001小于1的数据保留3位小数

def get_multiple_data(dict_data,dataset_name, variable_name, grid_x, grid_y,grid_concentration,file_name=''):   
    '''
    description: 
    param {dictionary} dict_data,key=dataset_name,value=variable_name,grid_x,grid_y,grid_concentration,file_name,默认为None，会创建一个新的字典
    param {str} dataset_name，数据集名称
    param {str} variable_name, 变量名
    param {numpy array(2D)} grid_x，经度坐标
    param {numpy array(2D)} grid_y，纬度坐标
    param {numpy array(2D)} grid_concentration，浓度值
    param {str} file_name, 数据集所在文件名，default=''
    return {dictionary} dict_data
    '''
    if dict_data is None:
        dict_data = {}
    if dataset_name in dict_data:
        print(f"dataset_name {dataset_name} already exists in dict_data")   
    dict_data[dataset_name] = {
        "file_name": file_name,
        "variable_name": variable_name,
        "grid_x": grid_x,
        "grid_y": grid_y,
        "grid_concentration": grid_concentration
    }    
    return dict_data

def show_multi_maps(dict_data,unit='',cmap='jet', show_lonlat=False,projection=None, boundary_file='',x_title='Longitude', y_title='Latitude',show_minmax=True,
                    default_min_value=-1,default_max_value=-1,panel_layout=None, show_original_grid=False,sharex=True, sharey=True, show_sup_title=False,is_wrf_out_data=False
                    ,points_data=None,fig_size=None,showplot=True,show_grid_line=True,value_format=None):
    '''
    description: 
    @param {dictionary} dict_data,key=dataset_name,value=variable_name,grid_x,grid_y,grid_concentration,file_name,默认为None，会创建一个新的字典
    param {str} dataset_name，数据集名称
    param {str} variable_name, 变量名
    param {numpy array(2D)} grid_x，经度坐标
    param {numpy array(2D)} grid_y，纬度坐标
    param {numpy array(2D)} grid_concentration，浓度值
    param {str} file_name, 数据集所在文件名，default=
    @param {str} unit, 单位
    @param {str} cmap, 颜色映射, default='jet'
    @param {str} projection, 地图投影, default='', 为空时使用PlateCarree投影
    @param {str} boundary_file, 行政边界文件, default='', 为空时不加载行政边界数据
    @param {str} x_title, x轴标题, default='Longitude'
    @param {str} y_title, y轴标题, default='Latitude'
    @param {bool} show_minmax, 是否显示最小最大值, default=True
    @param {float} default_min_value, 最小值, default=-1
    @param {float} default_max_value, 最大值, default=-1
    @param {str} panel_layout, 子图布局, default=None, 为空时自动计算布局
    @param {bool} show_original_grid, 是否显示原始网格, default=False
    @param {bool} sharex, 是否共享x轴, default=True
    @param {bool} sharey, 是否共享y轴, default=True
    @param {bool} show_sup_title, 是否显示共享XY轴标题, default=False
    @param {bool} show_lonlat, 是否显示经纬度
    @param {bool} showplot, 是否显示图形, default=True
    @param {bool} points_data, 是否显示点数据, default=None,示例：{'lon':df_monitor['lon'],'lat':df_monitor['lat'],'value':df_monitor['Monitor_Conc']}
    @param {bool} is_wrf_out_data, 是否是wrf输出数据, default=False
    @param {tuple} fig_size, 图形大小, default=None,示例：width,height=(6,6)
    @param {bool} show_grid_line, 是否显示网格线, default=True
    @param {str} value_format, 数值格式, default=None,示例：'.2f':保留2位小数；'.2e' ：采用科学计数法显示，并保留2位小数
    @return {None or matplotlib.figure} fig
    '''
    # print("")
    # 设置中文字体为系统中已安装的字体，如SimSun（宋体）、SimHei（黑体）
    #plt.rcParams['font.sans-serif'] = ['SimSun']  # 设置中文字体为宋体
    # 设置字体为 Times New Roman
    #plt.rcParams['axes.unicode_minus'] = False  # 用于正常显示负号
    # 创建可视化图
    data_types=dict_data.keys()
    case_num=len(data_types)  
    plot_columns,plot_rows=get_layout_col_row(case_num, panel_layout=panel_layout)
    if projection==None:
        projection = ccrs.PlateCarree()    
    origin_projection = ccrs.PlateCarree()  if is_wrf_out_data else projection 
    width,height =6,6 
    first_key = next(iter(dict_data))    
    height =math.ceil(width*dict_data[first_key]["grid_y"].shape[0]/dict_data[first_key]["grid_x"].shape[0]) if not isinstance( dict_data[first_key]["grid_y"] ,list) else width
    if fig_size!=None:       
        width,height=fig_size
    fig, axs = plt.subplots(plot_rows, plot_columns, figsize=(width * plot_columns, height* plot_rows),
                            subplot_kw={'projection': projection},sharex=sharex, sharey=sharey) 
    if boundary_file:
        reader = Reader(boundary_file)  
    if isinstance(axs, cartopy.mpl.geoaxes.GeoAxesSubplot):
        axs = np.array([axs])
    if show_sup_title==False and plot_rows>1:
        y_titles = [y_title] *plot_rows
        for ax, row in zip(axs[:,0], y_titles):
            ax.set_ylabel(row, rotation=90, size=10)
    # cols = ["LR","SVR","GBRT","DNN"]
    # for ax, col in zip(axs[0,:], cols,):
    #     ax.set_xlabel(col,size=10)
    axs = axs.ravel()  
    # all_grids=np.array([value["grid_concentration"] for data_type,value in dict_data.items()])
    # vmax_all=np.nanpercentile(all_grids,99.5)
    # vmin_all=np.nanpercentile(all_grids,0.5)  
    
    for ax,data_type in zip(axs,data_types):
        dic_sub_data = dict_data[data_type]
        file_name, variable_name, x, y, grid_concentration = dic_sub_data["file_name"], dic_sub_data["variable_name"], dic_sub_data["grid_x"], dic_sub_data["grid_y"], dic_sub_data["grid_concentration"]
        min_value, max_value, mean_value, total_value = np.nanmin(grid_concentration), np.nanmax(grid_concentration), np.nanmean(grid_concentration), np.nansum(grid_concentration)
        ax.text(0.5, 1.07, f'{data_type} {variable_name}', transform=ax.transAxes, fontsize=14, fontweight='bold',ha='center')

        vmax = np.nanpercentile(grid_concentration,99.5) if default_max_value == -1 else default_max_value #np.nanmax(grid_concentration)
        vmin =  np.nanpercentile(grid_concentration,0.5)  if default_min_value == -1 else default_min_value#np.nanmin(grid_concentration)
        if show_original_grid:
            contour = ax.pcolormesh(x, y, grid_concentration, cmap=cmap, vmin=vmin, vmax=vmax,transform=origin_projection)
            # # 绘制填色图
            # contour = ax.imshow(grid_concentration, origin='lower')
        else:
            contour = ax.contourf(x, y, grid_concentration, cmap=cmap,transform=origin_projection, vmin=vmin, vmax=vmax)
        if points_data is not None:  
            ax.scatter(x=points_data["lon"],y=points_data["lat"],s=50,c=points_data["value"],cmap=cmap,transform=origin_projection,vmin=vmin, vmax=vmax, edgecolor='black')      
        # 加载行政边界数据（示例使用shapefile文件）
        if boundary_file:                            
            geometries = reader.geometries()
            enshicity = cfeature.ShapelyFeature(geometries, origin_projection, edgecolor='k',facecolor='none')  # facecolor='none',设置面和线
            #enshicity = cfeature.ShapelyFeature(geometries, ccrs.PlateCarree(), edgecolor='k',facecolor='none')  # facecolor='none',设置面和线
            ax.add_feature(enshicity, linewidth=0.3)  # 添加市界细节  
        else:
            # 添加地图特征
            ax.add_feature(cfeature.COASTLINE, facecolor='none')# '-' 或 'solid'：实线# '--' 或 'dashed'：虚线# ':' 或 'dotted'：点线# '-.' 或 'dashdot'：点划线            
            ax.add_feature(cfeature.BORDERS, linestyle='solid', facecolor='none')
            ax.add_feature(cfeature.LAND, edgecolor='black',facecolor='none')  # facecolor='none'表示边界线围起来区域不填充颜色，只绘制边界线；
            ax.add_feature(cfeature.OCEAN, edgecolor='black', facecolor='none')
        min_longitude ,max_longitude,min_latitude,max_latitude=  round(x.min(),1),round(x.max(),1), round(y.min(),1), round(y.max(),1)       
        mean_or_total='Total' if 'ton' in unit else 'Mean'
        value=total_value if 'ton' in unit else mean_value 
        if is_wrf_out_data:            
            mesh = ax.gridlines(draw_labels=True, linestyle='--', linewidth=0.6, alpha=0.5, x_inline=False, y_inline=False, color='k',visible=show_grid_line)
            mesh.top_labels=False
            mesh.right_labels=False
            mesh.xformatter = LONGITUDE_FORMATTER
            mesh.yformatter = LATITUDE_FORMATTER  
            interval_x =math.ceil((max_longitude - min_longitude) / 10)
            interval_y =math.ceil((max_latitude - min_latitude) / 10)
            mesh.xlocator = mticker.FixedLocator(np.arange(min_longitude,max_longitude,interval_x))
            mesh.ylocator = mticker.FixedLocator(np.arange(min_latitude,max_latitude,interval_y))
            mesh.xlabel_style={'size':10}
            mesh.ylabel_style={'size':10}  
            if show_minmax:                   
                if value_format is not None:
                    # min_max_info= 'Min= {:.2f}, Max= {:.2f}, {:}={:.2f}'.format(min_value, max_value,mean_or_total,value)
                    min_max_info='Min= {}, Max= {}, {}={}'.format(format(min_value, value_format),format(max_value, value_format),mean_or_total,format(value, value_format))
                else:
                    if (mean_value<1000 and mean_value>0.001):                    
                        min_max_info= 'Min= {:.2f}, Max= {:.2f}, {:}={:.2f}'.format(min_value, max_value,mean_or_total,value)
                    else:
                        min_max_info= 'Min= {:.1e}, Max= {:.1e}, {:}={:.1e}'.format(min_value, max_value,mean_or_total,value)              
                ax.text(0.5, 1.02,min_max_info ,transform=ax.transAxes, fontsize=10, fontweight='bold',ha='center') 
        else :                   
            # 设置y轴范围
            ax.set_ylim(min_latitude, max_latitude)
            #print(min(y), max(y))
            # 设置 x 和 y 轴的 ticks 范围
            x_ticks,y_ticks = [],[]
            if max_longitude > 180 and show_lonlat:
                x_ticks = [-180, -120, -60, 0, 60, 120, 180]
            else:
                interval_x =round(float((max_longitude - min_longitude) / 7),2)
                x_ticks =np.round(np.arange(min_longitude, max_longitude + 0.1, interval_x),1)

            interval_y =round(float((max_latitude - min_latitude) / 7),2)
            y_ticks =np.round(np.arange(min_latitude, max_latitude + 0.1, interval_y),1)
            ax.set_xticks(x_ticks)  # 设置 x 轴的 ticks 范围
            ax.set_yticks(y_ticks)  # 设置 y 轴的 ticks 范围         
            # min_value, max_value, mean_value, total_value = np.nanmin(grid_concentration), np.nanmax(grid_concentration), np.nanmean(grid_concentration), np.nansum(grid_concentration)        
            if show_minmax:      
                if value_format is None: 
                    if unit=='ton':  
                        ax.set_xlabel(f"{ '' if show_sup_title else  x_title}\n" + 'Min= {:.2f}, Max= {:.2f}, Total={:.2f}'.format(min_value, max_value,value))
                    else:
                        #min_value, max_value, mean_value 采用科学计数法表示，保留1位小数           
                        ax.set_xlabel(f"{'' if show_sup_title else  x_title}\n"+'Min= {:.1e}, Max= {:.1e}, Mean={:.1e}'.format(min_value, max_value,value))
                else:                    
                    min_max_info='Min= {}, Max= {}, {}={}'.format(format(min_value, value_format),format(max_value, value_format),mean_or_total,format(value, value_format))
                    ax.set_xlabel(f"{'' if show_sup_title else  x_title}\n"+min_max_info)
            else:
                plt.xlabel(x_title)    

    if show_sup_title:
        fig.supxlabel(f'{x_title}',y=0.08,fontsize=15,fontweight="normal")#标签相对于图形的x位置，范围为0到1，默认为0.01，距离底部的距离为0.01，表示留出一小段空白。1表示距离顶部的距离为0
        fig.supylabel(f'{y_title}',x=0.08,fontsize=15,fontweight="normal")  
    plt.subplots_adjust(hspace=0.3)  # 调整子图之间的垂直间距
    # 添加颜色条
    cbar = plt.colorbar(contour, fraction=0.02, pad=0.04, label=f'{unit})',
                        ax=axs, orientation='vertical', shrink=0.7)  # 设置图例高度与图像高度相同, orientation='vertical', shrink=0.7

    cbar.set_label(f'({unit})', fontweight='bold')  # 设置标签字体加粗
   
    if showplot:       
        # 显示图形
        plt.show()
        return fig
    else:
        return fig

def show_delta_maps(dict_data,unit='',cmap='jet', show_lonlat=False,projection=None, boundary_file='',x_title='Longitude', y_title='Latitude',show_minmax=True,
                    default_min_value=-1,default_max_value=-1,panel_layout=None, show_original_grid=False,sharex=True, sharey=True, show_sup_title=False,is_wrf_out_data=False
                    ,points_data=None,delta_map_settings=None,show_dependenct_colorbar=False,showplot=True):
    '''
    description: 
    @param {dictionary} dict_data,key=dataset_name,value=variable_name,grid_x,grid_y,grid_concentration,file_name,默认为None，会创建一个新的字典
    param {str} dataset_name，数据集名称
    param {str} variable_name, 变量名
    param {numpy array(2D)} grid_x，经度坐标
    param {numpy array(2D)} grid_y，纬度坐标
    param {numpy array(2D)} grid_concentration，浓度值
    param {str} file_name, 数据集所在文件名，default=
    @param {str} unit, 单位
    @param {str} cmap, 颜色映射, default='jet'
    @param {str} projection, 地图投影, default='', 为空时使用PlateCarree投影
    @param {str} boundary_file, 行政边界文件, default='', 为空时不加载行政边界数据
    @param {str} x_title, x轴标题, default='Longitude'
    @param {str} y_title, y轴标题, default='Latitude'
    @param {bool} show_minmax, 是否显示最小最大值, default=True
    @param {float} default_min_value, 最小值, default=-1
    @param {float} default_max_value, 最大值, default=-1
    @param {str} panel_layout, 子图布局, default=None, 为空时自动计算布局
    @param {bool} show_original_grid, 是否显示原始网格, default=False
    @param {bool} sharex, 是否共享x轴, default=True
    @param {bool} sharey, 是否共享y轴, default=True
    @param {bool} show_sup_title, 是否显示共享XY轴标题, default=False
    @param {bool} show_lonlat, 是否显示经纬度
    @param {bool} showplot, 是否显示图形, default=True
    @param {bool} points_data, 是否显示点数据, default=None,示例：{'lon':df_monitor['lon'],'lat':df_monitor['lat'],'value':df_monitor['Monitor_Conc']}
    @param {bool} is_wrf_out_data, 是否是wrf输出数据, default=False
    @param {numpy array(2D)} delta_map_settings, 差值图的参数设定，default=None,示例：{'cmap':'coolwarm','default_min_value':-1,'default_max_value':-1}
    @return {None or matplotlib.figure} fig
    '''
    # print("")
    # 设置中文字体为系统中已安装的字体，如SimSun（宋体）、SimHei（黑体）
    #plt.rcParams['font.sans-serif'] = ['Arial']  # 设置中文字体为宋体
    # 设置字体为 Times New Roman
    #plt.rcParams['axes.unicode_minus'] = False  # 用于正常显示负号
    # 创建可视化图
    data_types=dict_data.keys()
    case_num=len(data_types)  
    plot_columns,plot_rows=get_layout_col_row(case_num, panel_layout=panel_layout)
    if projection==None:
        projection = ccrs.PlateCarree()    
    origin_projection = ccrs.PlateCarree()  if is_wrf_out_data else projection 
    width = 6
    first_key = next(iter(dict_data))    
    height =math.ceil(width*dict_data[first_key]["grid_y"].shape[0]/dict_data[first_key]["grid_x"].shape[0]) if not isinstance( dict_data[first_key]["grid_y"] ,list) else width
    fig, axs = plt.subplots(plot_rows, plot_columns, figsize=(width * plot_columns, height* plot_rows),
                            subplot_kw={'projection': projection},sharex=sharex, sharey=sharey) 
    if boundary_file:
        reader = Reader(boundary_file)  
    if isinstance(axs, cartopy.mpl.geoaxes.GeoAxesSubplot):
        axs = np.array([axs])
    if show_sup_title==False and plot_rows>1:
        y_titles = [y_title] *plot_rows
        for ax, row in zip(axs[:,0], y_titles):
            ax.set_ylabel(row, rotation=90, size=10)
    # cols = ["LR","SVR","GBRT","DNN"]
    # for ax, col in zip(axs[0,:], cols,):
    #     ax.set_xlabel(col,size=10)
    axs = axs.ravel()  
    for ax,data_type in zip(axs,data_types):
        dic_sub_data = dict_data[data_type]
        file_name, variable_name, x, y, grid_concentration = dic_sub_data["file_name"], dic_sub_data["variable_name"], dic_sub_data["grid_x"], dic_sub_data["grid_y"], dic_sub_data["grid_concentration"]
        min_value, max_value, mean_value, total_value = np.nanmin(grid_concentration), np.nanmax(grid_concentration), np.nanmean(grid_concentration), np.nansum(grid_concentration)
        ax.text(0.5, 1.07, f'{data_type} {variable_name}', transform=ax.transAxes, fontsize=14, fontweight='bold',ha='center')

        vmax = np.nanpercentile(grid_concentration,99.5) if default_max_value == -1 else default_max_value
        vmin = np.nanpercentile(grid_concentration,0.5) if default_min_value == -1 else default_min_value
        if 'delta' in str.lower(data_type) or 'delta' in str.lower(file_name):
            cmap_delta,default_delta_vmin,default_delta_vmax= delta_map_settings["cmap"],delta_map_settings["default_min_value"],delta_map_settings["default_max_value"]
            vmax_delta =np.nanpercentile(grid_concentration,99.5) if default_delta_vmax == -1 else default_delta_vmax
            vmin_delta =np.nanpercentile(grid_concentration,0.5) if default_delta_vmin == -1 else default_delta_vmin           
            max_value_delta=np.max([abs(vmax_delta),abs(vmin_delta)])            
            vmin_delta,vmax_delta=-max_value_delta,max_value_delta
            if show_original_grid:                
                contour = ax.pcolormesh(x, y, grid_concentration, cmap=cmap_delta, vmin=vmin_delta, vmax=vmax_delta,transform=origin_projection)           
            else:
                contour = ax.contourf(x, y, grid_concentration, cmap=cmap_delta,transform=origin_projection,  vmin=vmin_delta, vmax=vmax_delta)
            # 添加colorbar到当前子图
            cbar = plt.colorbar(contour, ax=ax,shrink=0.6)
            cbar.set_label(f'({unit})', fontweight='bold')  # 设置标签字体加粗
        else:
            if show_original_grid:
                contour = ax.pcolormesh(x, y, grid_concentration, cmap=cmap, vmin=vmin, vmax=vmax,transform=origin_projection)              
            else:
                contour = ax.contourf(x, y, grid_concentration, cmap=cmap,transform=origin_projection, vmin=vmin, vmax=vmax)
            if show_dependenct_colorbar:
                cbar = plt.colorbar(contour, ax=ax,shrink=0.6)
                cbar.set_label(f'({unit})', fontweight='bold')  # 设置标签字体加粗
        if points_data is not None:  
            ax.scatter(x=points_data["lon"],y=points_data["lat"],s=50,c=points_data["value"],cmap=cmap,transform=origin_projection,vmin=vmin, vmax=vmax, edgecolor='black')      
        # 加载行政边界数据（示例使用shapefile文件）
        if boundary_file:                            
            geometries = reader.geometries()
            enshicity = cfeature.ShapelyFeature(geometries, origin_projection, edgecolor='k',facecolor='none')  # facecolor='none',设置面和线
            #enshicity = cfeature.ShapelyFeature(geometries, ccrs.PlateCarree(), edgecolor='k',facecolor='none')  # facecolor='none',设置面和线
            ax.add_feature(enshicity, linewidth=0.3)  # 添加市界细节  
        else:
            # 添加地图特征
            ax.add_feature(cfeature.COASTLINE, facecolor='none')# '-' 或 'solid'：实线# '--' 或 'dashed'：虚线# ':' 或 'dotted'：点线# '-.' 或 'dashdot'：点划线            
            ax.add_feature(cfeature.BORDERS, linestyle='solid', facecolor='none')
            ax.add_feature(cfeature.LAND, edgecolor='black',facecolor='none')  # facecolor='none'表示边界线围起来区域不填充颜色，只绘制边界线；
            ax.add_feature(cfeature.OCEAN, edgecolor='black', facecolor='none')
        min_longitude ,max_longitude,min_latitude,max_latitude=  round(x.min(),1),round(x.max(),1), round(y.min(),1), round(y.max(),1)       
        if is_wrf_out_data:            
            mesh = ax.gridlines(draw_labels=True, linestyle='--', linewidth=0.6, alpha=0.5, x_inline=False, y_inline=False, color='k')
            mesh.top_labels=False
            mesh.right_labels=False
            mesh.xformatter = LONGITUDE_FORMATTER
            mesh.yformatter = LATITUDE_FORMATTER  
            # mesh.xlocator = mticker.FixedLocator(np.arange(min_longitude,max_longitude,1))
            # mesh.ylocator = mticker.FixedLocator(np.arange(min_latitude,max_latitude,1))
            interval_x =math.ceil((max_longitude - min_longitude) / 10)
            interval_y =math.ceil((max_latitude - min_latitude) / 10)
            mesh.xlocator = mticker.FixedLocator(np.arange(min_longitude,max_longitude,interval_x))
            mesh.ylocator = mticker.FixedLocator(np.arange(min_latitude,max_latitude,interval_y))
            mesh.xlabel_style={'size':10}
            mesh.ylabel_style={'size':10}  
            if show_minmax:
                mean_or_total='Total' if 'ton' in unit else 'Mean'
                value=total_value if 'ton' in unit else mean_value            
                if (abs(mean_value)<2000 and abs(mean_value)>0.001):
                    min_max_info= 'Min= {:.2f}, Max= {:.2f}, {:}={:.2f}'.format(min_value, max_value,mean_or_total,value)
                else:
                    min_max_info= 'Min= {:.1e}, Max= {:.1e}, {:}={:.1e}'.format(min_value, max_value,mean_or_total,value)              
                ax.text(0.5, 1.02,min_max_info ,transform=ax.transAxes, fontsize=9,ha='center') #, fontweight='bold'
        else :                   
            # 设置y轴范围
            ax.set_ylim(min_latitude, max_latitude)
            #print(min(y), max(y))
            # 设置 x 和 y 轴的 ticks 范围
            x_ticks,y_ticks = [],[]
            if max_longitude > 180 and show_lonlat:
                x_ticks = [-180, -120, -60, 0, 60, 120, 180]
            else:
                interval_x =round(float((max_longitude - min_longitude) / 7),2)
                x_ticks =np.round(np.arange(min_longitude, max_longitude + 0.1, interval_x),1)

            interval_y =round(float((max_latitude - min_latitude) / 7),2)
            y_ticks =np.round(np.arange(min_latitude, max_latitude + 0.1, interval_y),1)
            ax.set_xticks(x_ticks)  # 设置 x 轴的 ticks 范围
            ax.set_yticks(y_ticks)  # 设置 y 轴的 ticks 范围         
            # min_value, max_value, mean_value, total_value = np.nanmin(grid_concentration), np.nanmax(grid_concentration), np.nanmean(grid_concentration), np.nansum(grid_concentration)        
            if show_minmax:                
                if unit=='ton':              
                    ax.set_xlabel(f"{ '' if show_sup_title else  x_title}\n" + 'Min= {:.2f}, Max= {:.2f}, Total={:.2f}'.format(min_value, max_value,total_value))
                else:
                    #min_value, max_value, mean_value 采用科学计数法表示，保留1位小数           
                    ax.set_xlabel(f"{'' if show_sup_title else  x_title}\n"+'Min= {:.1e}, Max= {:.1e}, Mean={:.1e}'.format(min_value, max_value,mean_value))
            else:
                plt.xlabel(x_title)    

    if show_sup_title:
        fig.supxlabel(f'{x_title}',y=0.08,fontsize=15,fontweight="normal")#标签相对于图形的x位置，范围为0到1，默认为0.01，距离底部的距离为0.01，表示留出一小段空白。1表示距离顶部的距离为0
        fig.supylabel(f'{y_title}',x=0.08,fontsize=15,fontweight="normal")  
    plt.subplots_adjust(hspace=0.3)  # 调整子图之间的垂直间距
    if not show_dependenct_colorbar and 'delta' not in data_type:
        # 添加颜色条
        cbar = plt.colorbar(contour, fraction=0.02, pad=0.04, label=f'{unit})',
                            ax=axs, orientation='vertical', shrink=0.7)  # 设置图例高度与图像高度相同, orientation='vertical', shrink=0.7

        cbar.set_label(f'({unit})', fontweight='bold')  # 设置标签字体加粗
   
    if showplot:       
        # 显示图形
        plt.show()
        return fig
    else:
        return fig

def show_single_map(grid_x, grid_y, grid_concentration, title='', x_title='Longitude', y_title='Latitude', unit='', boundary_file='',cmap='jet',figsize=None, show_minmax=True,default_min_value=-1,default_max_value=-1, show_original_grid=False,showplot=True):
    '''
     :param grid_x:list of ndarrays, 通常np.meshgrid后结果
     :param grid_y:list of ndarrays, 通常np.meshgrid后结果
     :param grid_concentration:list of ndarrays，需要展示的网格化数据
     :param title:标题名称
     :param x_title:x轴标题
     :param y_title:y轴标题
     :param unit:数据单位
     :param boundary_file:边界线路径，如设定路径，则会在现有数据基础上追加边界线
     :param show_minmax:是否显示最大最小值，默认显示
     :param showplot:是否绘图，默认绘图；False将返回fig,用于保存
     :return:showplot=True，直接绘图，没有返回值。showplot=False,返回图表数据变量，可直接调用fig.savefig(output_file, dpi=600)进行保存图片。
    '''
    # 创建可视化图
    fig = plt.figure(figsize=figsize)#figsize=(10, 5)
    # 创建一个地图投影
    ax = plt.axes(projection=ccrs.PlateCarree())
    title_text = ax.text(0.5, 1.05, f'{title} ', transform=ax.transAxes, fontsize=12, ha='center')

    # 添加地图特征
    ax.add_feature(cfeature.COASTLINE, facecolor='none')
    # '-' 或 'solid'：实线    #'--' 或 'dashed'：虚线    #':' 或 'dotted'：点线    #'-.' 或 'dashdot'：点划线
    ax.add_feature(cfeature.BORDERS, linestyle='solid', facecolor='none')
    ax.add_feature(cfeature.LAND, edgecolor='black', facecolor='none')  # facecolor='none'表示边界线围起来区域不填充颜色，只绘制边界线；
    ax.add_feature(cfeature.OCEAN, edgecolor='black', facecolor='none')

    max_value = np.nanmax(grid_concentration) if default_max_value==-1 else default_max_value
    min_value = np.nanmin(grid_concentration) if default_min_value==-1 else default_min_value
    
    # data_subset.lon 和 data_subset.lat 是一维数组，data_subset 是对应的二维浓度数据
    # contour = ax.contourf(data_subset.lon, data_subset.lat, data_subset, cmap='jet', transform=ccrs.PlateCarree())#展示选定范围cmap='viridis',coolwarm
    # contour = ax.contourf(grid_x, grid_y, grid_concentration, cmap='viridis')  # 使用插值后的数据绘制等高线填充图
    if show_original_grid:
        contour =ax.pcolormesh(grid_x, grid_y, grid_concentration,cmap=cmap,vmin=min_value, vmax=max_value)
    else:        
        contour = ax.contourf(grid_x, grid_y, grid_concentration, cmap=cmap,
                              transform=ccrs.PlateCarree(), vmin=min_value, vmax=max_value)  # 展示选定范围cmap='viridis',coolwarm
    
    # 加载行政边界数据（示例使用shapefile文件）
    if boundary_file:
        boundaries = gpd.read_file(boundary_file)
        boundaries.boundary.plot(color='black', linewidth=0.8, ax=plt.gca())  # 添加行政边界线

    # 添加颜色条
    if unit:unit=f'({unit})'
    cbar = plt.colorbar(contour, fraction=0.02, pad=0.04, label=f'{unit}',
                        ax=ax)  # 设置图例高度与图像高度相同, orientation='vertical', shrink=0.7

    # 设置y轴范围
    ax.set_ylim(grid_y.min(), grid_y.max())#max(values) 是 Python 内置的函数，用于找到给定可迭代对象（例如列表、元组等）中的最大值。values.max() 是 NumPy 数组对象的方法，用于计算 NumPy 数组中元素的最大值。
    max_longitude=grid_x.max()
    min_longitude=grid_x.min()
    max_latitude=grid_y.max()
    min_latitude =grid_y.min()
    # 设置 x 和 y 轴的 ticks 范围
    x_ticks = [-180, -120, -60, 0, 60, 120, 180]
    y_ticks = [-90, -60, -30, 0, 30, 60, 90]
    if max_longitude > 180:
        x_ticks = [-180, -120, -60, 0, 60, 120, 180]
    else:
        interval_x = int((max_longitude - min_longitude) / 7)
        x_ticks = np.arange(min_longitude, max_longitude + 0.1*interval_x, interval_x)

    interval_y = int((max_latitude - min_latitude) / 7)
    y_ticks = np.arange(min_latitude, max_latitude + 0.1*interval_y, interval_y)
    ax.set_xticks(x_ticks)  # 设置 x 轴的 ticks 范围
    ax.set_yticks(y_ticks)  # 设置 y 轴的 ticks 范围

    # 添加标题和标签
    plt.ylabel(y_title)
    # 在x轴标题下方添加一行显示最大值和最小值
    if show_minmax:
        plt.xlabel(x_title+f"\nMin= {format_data(min_value)}, Max= {format_data(np.nanmax(grid_concentration))}, Total= {format_data(np.nansum(grid_concentration))}")
        print(title+f"\nMin= {format_data(min_value)}, Max= {format_data(np.nanmax(grid_concentration))}, Total= {format_data(np.nansum(grid_concentration))}")
    else:
        plt.xlabel(x_title)
    if showplot:
        # 显示图形
        plt.show()
    else:
        return fig

def plot_cmaq_concentration_maps(dic_grid_values, model_property,unit='',cmap='jet', show_lonlat=False, boundary_file='',x_title='Longitude', y_title='Latitude',show_minmax=True,default_min_value=-1,default_max_value=-1, panel_layout=PaneLayout.SquareColPreferred, show_original_grid=False, showplot=True):
    '''
    @description: 绘制CMAQ浓度图
    @param {type} dic_grid_values: 字典，包含不同时间的不同变量的网格化数据
    @param {type} model_property: 模型属性对象，包含模型的属性信息，如行数、列数、经纬度范围等
    @param {type} unit: 单位
    @param {type} cmap: 颜色映射
    @param {type} show_lonlat: 是否显示经纬度
    @param {type} boundary_file: 边界文件路径
    @param {type} x_title: x轴标题
    @param {type} y_title: y轴标题
    @param {type} show_minmax: 是否显示最大最小值
    @param {type} default_min_value: 默认最小值
    @param {type} default_max_value: 默认最大值
    @param {type} panel_layout: 子图布局
    @param {type} show_original_grid: 是否显示原始网格
    @param {type} showplot: 是否显示图形
    @return: fig
    '''
    cols,rows = model_property.cols,model_property.rows
    x,y = [],[]
    # x_lable,y_lable='Longitude','Latitude'
    projection = ccrs.PlateCarree()
    if show_lonlat:
        lon_start, lat_start, lon_end, lat_end = model_property.lon_start,model_property.lat_start,model_property.lon_end,model_property.lat_end
        x = np.linspace(lon_start, lon_end, cols)
        y = np.linspace(lat_start, lat_end, rows)
    else:
        x = np.linspace(1, cols, cols)
        y = np.linspace(1, rows, rows)
        # x_lable, y_lable = 'Column', 'Row'
        # 创建 Lambert 投影对象
        #projection = ccrs.LambertConformal(central_longitude=112, central_latitude=30,standard_parallels=(25, 40))
        projection =model_property.projection #ccrs.CRS(model_property.proj4_string)# ccrs.Projection(model_property.proj4_string)

    # Define the latitude and longitude ranges you want to display
    min_longitude = x.min()
    max_longitude = x.max()
    min_latitude = y.min()
    max_latitude = y.max()
    # 定义要显示的经纬度范围
    lon_range = slice(min_longitude, max_longitude)  # 替换为您的经度范围
    lat_range = slice(min_latitude, max_latitude)  # 替换为您的纬度范围
    # 设置中文字体为系统中已安装的字体，如SimSun（宋体）、SimHei（黑体）
    plt.rcParams['font.sans-serif'] = ['SimSun']  # 设置中文字体为宋体
    # 设置字体为 Times New Roman
    plt.rcParams['axes.unicode_minus'] = False  # 用于正常显示负号
    # 创建可视化图
    case_list = list(dic_grid_values.keys())
    plot_rows = len(case_list)
    plot_columns = len(dic_grid_values[case_list[0]])    
    
    # fig, axs = plt.figure(1,columns,figsize=(12, 5))
    fig, axs = plt.subplots(plot_rows, plot_columns, figsize=(6 * plot_columns, 6 * plot_rows),
                            subplot_kw={'projection': projection})

    if boundary_file:
        reader = Reader(boundary_file)
        #geometries = reader.geometries()
    row_index = 0
    for date, dic_data in dic_grid_values.items():
        col_index = 0
        for variable_name, data in dic_data.items():
            if plot_rows>1 and plot_columns>1:
                ax = axs[row_index, col_index]
            elif plot_rows==1 and plot_columns>1:
                ax = axs[col_index]
            else:
                ax = axs[row_index]               
       
            # Use .sel() to select the specific latitude and longitude range
            #data_subset =data# data.sel(lon=slice(min_longitude, max_longitude), lat=slice(min_latitude, max_latitude))
            # 创建一个地图投影
            # ax = plt.axes(projection=ccrs.PlateCarree())
            title_text = ax.text(0.5, 1.05, f'{date}-{variable_name} ', transform=ax.transAxes, fontsize=16, fontweight='bold',ha='center')

            vmax = np.nanmax(data) if default_max_value == -1 else default_max_value
            vmin = np.nanmin(data) if default_min_value == -1 else default_min_value
            if show_original_grid:
                contour = ax.pcolormesh(x, y, data, cmap=cmap, vmin=vmin, vmax=vmax,transform=projection)
            else:
                contour = ax.contourf(x, y, data, cmap=cmap,
                                      transform=projection, vmin=vmin, vmax=vmax)
                
            #contour = ax.contourf(x, y, data, cmap='jet',                                  transform=projection)  # 展示选定范围cmap='viridis',coolwarm
            # 加载行政边界数据（示例使用shapefile文件）
            if boundary_file:
                #reader = Reader(boundary_file)
                geometries = reader.geometries()
                enshicity = cfeature.ShapelyFeature(geometries, projection, edgecolor='k',
                                                 facecolor='none')  # facecolor='none',设置面和线
                ax.add_feature(enshicity, linewidth=0.3)  # 添加市界细节
            else:
                # 添加地图特征
                ax.add_feature(cfeature.COASTLINE, facecolor='none')
                # '-' 或 'solid'：实线            # '--' 或 'dashed'：虚线            # ':' 或 'dotted'：点线            # '-.' 或 'dashdot'：点划线
                ax.add_feature(cfeature.BORDERS, linestyle='solid', facecolor='none')
                ax.add_feature(cfeature.LAND, edgecolor='black',
                               facecolor='none')  # facecolor='none'表示边界线围起来区域不填充颜色，只绘制边界线；
                ax.add_feature(cfeature.OCEAN, edgecolor='black', facecolor='none')
            # 设置y轴范围
            ax.set_ylim(min(y), max(y))
            #print(min(y), max(y))
            # 设置 x 和 y 轴的 ticks 范围
            x_ticks,y_ticks = [],[]
            if max_longitude > 180 and show_lonlat:
                x_ticks = [-180, -120, -60, 0, 60, 120, 180]
            else:
                interval_x =round(float((max_longitude - min_longitude) / 7),2)
                x_ticks =np.round(np.arange(min_longitude, max_longitude + 0.1, interval_x),2)

            interval_y =round(float((max_latitude - min_latitude) / 7),2)
            y_ticks =np.round(np.arange(min_latitude, max_latitude + 0.1, interval_y),2)
            ax.set_xticks(x_ticks)  # 设置 x 轴的 ticks 范围
            ax.set_yticks(y_ticks)  # 设置 y 轴的 ticks 范围
            # 在x轴标题下方添加一行显示最大值和最小值
            max_value = data.max().values
            min_value = data.min().values
            
            if show_minmax:                
                if unit=='ton':
                    ax.set_xlabel(f'{x_title}' + '\n\nMin= {:.2f}, Max= {:.2f}, Total={:.2f}'.format(min_value, max_value,np.nansum(data.values)))
                else:
                    #min_value, max_value, mean_value 采用科学计数法表示，保留1位小数
                    ax.set_xlabel(f'{x_title}'+'\n\nMin= {:.1e}, Max= {:.1e}, Mean={:.1e}'.format(min_value, max_value,np.nanmean(data.values)))
            else:
                plt.xlabel(x_title)
            #ax.set_xlabel(f'{x_title}'+'\n\nMin= {:.2f}, Max= {:.2f}'.format(min_value, max_value), fontweight='bold')
            # 添加标题和标签
            if col_index == 0:
                ax.set_ylabel(f'{y_title}', fontweight='bold')
            col_index = col_index + 1
        row_index = row_index + 1
    # 添加颜色条
    cbar = plt.colorbar(contour, fraction=0.02, pad=0.04, label=f'{unit})',
                        ax=axs, orientation='vertical', shrink=0.7)  # 设置图例高度与图像高度相同, orientation='vertical', shrink=0.7

    cbar.set_label(f'({unit})', fontweight='bold')  # 设置标签字体加粗
    if showplot:
        # 显示图形
        plt.show()
        return fig
    else:
        return fig
    
def get_bc_whole_domain_data(bc_file,variable_name,time_name='TSTEP',layer_name='LAY',specified_time_step=None,specified_layer_step=None,show_lonlat=False):
    '''
    @description: 获取boundary condition文件中的全域数据
    @param {str} boundary_file, 行政边界文件
    @param {str} variable_name, 变量名
    @param {time_name} time_name, 时间变量名, default='TSTEP'
    @param {layer_name} layer_name, 层变量名, default='LAY'
    @param {int} specified_time_step, 指定时间步长索引, default=None，为None时获取所有时间步长
    @param {int} specified_layer_step, 指定层步长索引, default=None, 为None时获取所有层步长
    @param {bool} show_lonlat, 是否返回经纬度信息，默认为False
    @return {tuple} grid,x,y,projection,unit: 网格数据， x,y坐标,投影,单位
    '''
    from esil.RSM.Model_Property import model_attribute  
    import xarray as xr
    model = model_attribute(bc_file)
    ds = xr.open_dataset(bc_file, decode_times=False)
    if hasattr(ds[variable_name], 'units'):
        unit = ds[variable_name].units.strip()
    time_step_count = ds[time_name].size
    layer_count = ds[layer_name].size
    if specified_time_step is not None:
        time_step_count = 1
    if specified_layer_step is not None:
        layer_count = 1
    grid = np.full((time_step_count,layer_count,model.rows,model.cols), np.nan)
    dict_grid_index = {}
    for time_index in range(ds[time_name].size):
        if specified_time_step is not None and time_index != specified_time_step:
            continue        
        print(f"Processing time_index={time_index}")
        for layer_index in range(ds[layer_name].size): 
            if specified_layer_step is not None and layer_index != specified_layer_step:
                continue           
            for row_index,y in enumerate(model.y_coords):
                for col_index,x in enumerate(model.x_coords):                   
                    if dict_grid_index.get((row_index,col_index)) is None:   
                        idx = np.where((model.x_coords_bc == x) & (model.y_coords_bc == y))    # 搜索给定经纬度的索引                 
                        dict_grid_index[(row_index,col_index)] = idx
                    else:
                        idx = dict_grid_index.get((row_index,col_index))
                    if idx[0].size > 0:                        
                        # target_concentration = ds[variable_name][time_index,layer_index,:].values[idx]
                        grid[time_index,layer_index,row_index,col_index] = ds[variable_name][time_index,layer_index,:].values[idx]
                        #print(f'row_index={row_index},col_index={col_index},idx={idx}')
    x,y=model.get_xy_coordinates(show_lonlat=show_lonlat)                
    return grid,x,y,model.projection,unit

def plot_stacked_bar_on_map(data, title='', x_title='Longitude', y_title='Latitude', unit='', boundary_file='', cmap='jet', figsize=None, show_minmax=True, showplot=True):
    '''
    @description: 在地图上绘制堆叠条形图
    @param {dict} data, 数据字典，key=AreaID, value={label: value}
    @param {str} title, 标题
    @param {str} x_title, x轴标题
    @param {str} y_title, y轴标题
    @param {str} unit, 单位
    @param {str} boundary_file, 行政边界文件
    @param {str} cmap, 颜色映射
    @param {tuple} figsize, 图形尺寸
    @param {bool} show_minmax, 是否显示最小最大值
    @param {bool} showplot, 是否显示图形
    @return {matplotlib.figure.Figure} fig, 图形对象  
    '''
    # Load administrative boundary data
    gdf = gpd.read_file(boundary_file)

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(12, 8))

    # Plot administrative boundaries
    gdf.boundary.plot(ax=ax, linewidth=1.0, color='black')

    # Get centroids of each area
    gdf['centroid'] = gdf['geometry'].centroid

    # Define colors for labels
    label_color_dict = {}

    # Initialize bottom values for stacking bars
    bottom_values = {label: [0] * len(gdf) for label in data[list(data.keys())[0]]}

    # Iterate over each area to plot stacked bar charts
    for idx, row in gdf.iterrows():
        x, y = row['centroid'].x, row['centroid'].y
        key = row['AreaID']
        data_values = data.get(key, {})  # Use .get() method to handle missing keys gracefully
        if not data_values:
            continue
        for label, value in data_values.items():
            if label not in label_color_dict:
                label_color_dict[label] = next(ax._get_lines.prop_cycler)['color']
            ax.bar(key, value, color=label_color_dict[label], bottom=bottom_values[label], alpha=1)
            bottom_values[label] = [sum(x) for x in zip(bottom_values[label], [value])]

    # Create legend
    legend_handles = [plt.Line2D([0], [0], marker='o', color='w', markerfacecolor=color, label=label)
                      for label, color in label_color_dict.items()]
    plt.legend(handles=legend_handles, loc='upper right')

    # Set x and y axis limits
    ax.set_xlim(gdf.total_bounds[0], gdf.total_bounds[2])
    ax.set_ylim(gdf.total_bounds[1], gdf.total_bounds[3])

    # Set titles and labels
    ax.set_title(title)
    ax.set_xlabel(x_title)
    ax.set_ylabel(y_title)

    # Show plot if specified
    if showplot:
        plt.show()
        
def plot_bar_on_map(data, title='', x_title='Longitude', y_title='Latitude', unit='', boundary_file='', cmap='jet', figsize=None, show_minmax=True, showplot=True):
    '''
    @description: 在地图上绘制条形图
    @param {dict} data, 数据字典，key=AreaID, value={label: value}
    @param {str} title, 标题
    @param {str} x_title, x轴标题
    @param {str} y_title, y轴标题
    @param {str} unit, 单位
    @param {str} boundary_file, 行政边界文件
    @param {str} cmap, 颜色映射
    @param {tuple} figsize, 图形尺寸
    @param {bool} show_minmax, 是否显示最小最大值
    @param {bool} showplot, 是否显示图形
    @return {matplotlib.figure.Figure} fig, 图形对象     
    '''
    # Load administrative boundary data
    gdf = gpd.read_file(boundary_file)

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(12, 8))

    # Plot administrative boundaries
    gdf.boundary.plot(ax=ax, linewidth=1.0, color='black')

    # Get centroids of each area
    gdf['centroid'] = gdf['geometry'].centroid

    # Define colors for labels
    label_color_dict = {}

    # Iterate over each area to plot bar charts
    for idx, row in gdf.iterrows():
        x, y = row['centroid'].x, row['centroid'].y
        key = row['AreaID']
        data_values = data.get(key, {})  # Use .get() method to handle missing keys gracefully
        if not data_values:
            continue
        for label in data_values:
            if label not in label_color_dict:
                label_color_dict[label] = next(ax._get_lines.prop_cycler)['color']

        values = list(data_values.values())  # Values for bar heights
        labels = list(data_values.keys())   # Labels for bars
        colors = [label_color_dict[label] for label in labels]

        # Plot bar chart
        ax.bar(x,labels, values, color=colors, alpha=0.7)
    
    # Create legend
    legend_handles = [plt.Line2D([0], [0], marker='o', color='w', markerfacecolor=color, label=label)
                      for label, color in label_color_dict.items()]
    plt.legend(handles=legend_handles, loc='upper right')

    # Set x and y axis limits
    ax.set_xlim(gdf.total_bounds[0], gdf.total_bounds[2])
    ax.set_ylim(gdf.total_bounds[1], gdf.total_bounds[3])

    # Set titles and labels
    ax.set_title(title)
    ax.set_xlabel(x_title)
    ax.set_ylabel(y_title)

    # Show plot if specified
    if showplot:
        plt.show()
        
def plot_pie1_on_map(data, title='', x_title='Longitude', y_title='Latitude', unit='', boundary_file='', cmap='jet', figsize=None, show_minmax=True, showplot=True):
    '''
    @description: 在地图上绘制饼图
    @param {dict} data, 数据字典，key=AreaID, value={label: value}
    @param {str} title, 标题
    @param {str} x_title, x轴标题
    @param {str} y_title, y轴标题
    @param {str} unit, 单位
    @param {str} boundary_file, 行政边界文件
    @param {str} cmap, 颜色映射
    @param {tuple} figsize, 图形尺寸
    @param {bool} show_minmax, 是否显示最小最大值
    @param {bool} showplot, 是否显示图形
    @return {matplotlib.figure.Figure} fig, 图形对象     
    '''
    # Load administrative boundary data
    
    gdf = gpd.read_file(boundary_file)

    # Create a figure and axis
    fig, ax = plt.subplots(figsize=(12, 8))

    # Plot administrative boundaries
    gdf.boundary.plot(ax=ax, linewidth=1.0, color='black')

    # Get centroids of each area
    gdf['centroid'] = gdf['geometry'].centroid

    # Define colors for labels
    label_color_dict = {}

    # Iterate over each area to plot pie charts
    for idx, row in gdf.iterrows():
        x, y = row['centroid'].x, row['centroid'].y
        key = row['AreaID']
        data_values = data.get(key, {})  # Use .get() method to handle missing keys gracefully
        if not data_values:
            continue
        for label in data_values:
            if label not in label_color_dict:
                label_color_dict[label] = next(ax._get_lines.prop_cycler)['color']

        sizes = list(data_values.values())  # Values for pie chart slices
        labels = list(data_values.keys())   # Labels for pie chart slices
        colors = [label_color_dict[label] for label in labels]

        # Plot pie chart
        ax.pie(sizes, colors=colors, labels=None, startangle=90, radius=0.1, center=(x, y))

    # Create legend
    legend_handles = [plt.Line2D([0], [0], marker='o', color='w', markerfacecolor=color, label=label)
                      for label, color in label_color_dict.items()]
    plt.legend(handles=legend_handles, loc='upper right')

    # Set x and y axis limits
    ax.set_xlim(gdf.total_bounds[0], gdf.total_bounds[2])
    ax.set_ylim(gdf.total_bounds[1], gdf.total_bounds[3])

    # Set titles and labels
    ax.set_title(title)
    ax.set_xlabel(x_title)
    ax.set_ylabel(y_title)

    # Show plot if specified
    if showplot:
        plt.show()

def plot_pie_on_map(data,title='', x_title='Longitude', y_title='Latitude', unit='', boundary_file='',cmap='jet',figsize=None, show_minmax=True, showplot=True):
    '''
    @description: 在地图上绘制饼图
    @param {dict} data, 数据字典，key=AreaID, value={label: value}
    @param {str} title, 标题
    @param {str} x_title, x轴标题
    @param {str} y_title, y轴标题
    @param {str} unit, 单位
    @param {str} boundary_file, 行政边界文件
    @param {str} cmap, 颜色映射
    @param {tuple} figsize, 图形尺寸
    @param {bool} show_minmax, 是否显示最小最大值
    @param {bool} showplot, 是否显示图形
    @return {matplotlib.figure.Figure} fig, 图形对象    
    '''    
    # 加载行政边界数据（示例使用shapefile文件）
    gdf = gpd.read_file(boundary_file)
    # 创建一个图形和子图
    fig, ax = plt.subplots(figsize=(12, 8))
    # 去除行政区域的颜色，并保留边界线
    gdf.boundary.plot(ax=ax, linewidth=1.0, color='black')
    # 获取每个区域的几何中心坐标
    gdf['centroid'] = gdf['geometry'].centroid
    # 统计所有区域数据中出现的标签和它们的颜色
    label_color_dict = {}
    # 遍历每个区域的中心坐标，绘制饼图
    for idx, row in gdf.iterrows():
        x, y = row['centroid'].x, row['centroid'].y
        key=row['AreaID']
        data_values=data[key]
        #data_values = row['data_column']  # 从数据列中获取用于绘制饼图的数据
        if not data_values:  # 如果数据为空，跳过绘制饼图
            continue
        for label in data_values:
            if label not in label_color_dict:
                label_color_dict[label] = next(ax._get_lines.prop_cycler)['color']

    # 遍历每个区域的中心坐标，绘制饼图
    for idx, row in gdf.iterrows():
        x, y = row['centroid'].x, row['centroid'].y
        key = row['AreaID']
        data_values = data[key]
        if not data_values:  # 如果数据为空，跳过绘制饼图
            continue
        sizes = [value for value in data_values.values()]  # 饼图分块的大小
        labels = [key for key in data_values.keys()]  # 饼图分块的标签
        colors = [label_color_dict[label] for label in labels]  # 使用预先定义的颜色
        ax.pie(sizes, colors=colors, labels=None, startangle=90, radius=0.1, center=(x, y))

    # 创建图例
    legend_handles = [plt.Line2D([0], [0], marker='o', color='w', markerfacecolor=color, label=label)
                      for label, color in label_color_dict.items()]
    plt.legend(handles=legend_handles, loc='upper right')

    # 设置 x 和 y 轴的范围
    ax.set_xlim(gdf.total_bounds[0], gdf.total_bounds[2])
    ax.set_ylim(gdf.total_bounds[1], gdf.total_bounds[3])
    # 设置标题和标签等其他绘图参数
    ax.set_title('Pie Charts on Map')
    ax.set_xlabel('Longitude')
    ax.set_ylabel('Latitude')
   #ax.grid(True)

    plt.show()


if __name__=="__main__":
    import math
    import os
    from esil.file_helper import get_files
    from esil.netcdf_helper import get_extent_data
    from esil.earth_helper import get_m2
    from esil.RSM.Model_Property import model_attribute   
  

    extent =[]
    china_extent = [70.25, 136.25, 10.125, 55.25]     
    guangdong_extent=[109.5, 117.5, 19, 27]    
    #extent=guangdong_extent
    boundary_file=''    
    json_file=r'D:\Devin\MyTestDemo\PythonDemo\NetcdfReader\CO2 Visualization\100000.json' 
    unit='ton'
    unit='mol/m2/s'
    default_max_value=10*math.pow(10,6) if unit=='ton' else 30*math.pow(10,-7)
    default_min_value=0 if unit=='ton' else -30*math.pow(10,-7)    
    dict_data={}
    ###开始：绘制SCUT.GLOBAL.fluxBIO.BAYES数据的示例   
    
    # nc_file=r'E:\CO2\CarbonTracker\CT2022.flux1x1.202001.nc'   
    # variable_name = 'bio_posterior'  
    # longitude_name='lon'
    # latitude_name='lat'  
    # cmap='seismic'
    # # 获取目录中的所有文件
    # file_path=r'E:\CO2\data\emis\posterior emission\SCUT_GLOBAL_FLUX\ENKF_Conc'
    # files = get_files(file_path,recursive=False,)
    # variable_name='xco2' if 'Conc' in file_path else variable_name
    # default_min_value=-1 if 'Conc' in file_path else default_min_value
    # default_max_value=-1 if 'Conc' in file_path else default_max_value
    # for index,nc_file in enumerate(files):     
    #     temp,unit_origin=get_extent_data(nc_file, variable_name,x_name=longitude_name,y_name=latitude_name,extent=extent)
    #     x=temp[longitude_name].data #data.longitude.values
    #     y=temp[latitude_name].data      
    #     name=os.path.basename(nc_file).replace('SCUT.GLOBAL.fluxBIO.','')
    #     grid_x, grid_y=np.meshgrid(x,y)
    #     grid_area=get_m2(grid_x, grid_y) 
    #     if 'Conc'in file_path:
    #         data=temp*math.pow(10,6)# μmol/mol = ppmv
    #         cmap='jet'
    #         unit='ppm'
    #     else:
    #         data=temp*((44.0/12)/math.pow(10,6))*grid_area*12 if unit=='ton' else temp*1/44*(44.0/12)/(24*3600*30)
    #     dict_data=get_multiple_data(dict_data,name,variable_name, x, y,data,nc_file)
    #     # if index==0:
    #     #     break
      
    # show_multi_maps(dict_data,show_original_grid=True,default_max_value=default_max_value,default_min_value=default_min_value,boundary_file=boundary_file,unit=unit,cmap=cmap)   
   
    ###结束：绘制SCUT.GLOBAL.fluxBIO.BAYES数据的示例  
    
    ###开始：绘制MEIC_3km数据的示例  
    import xarray as xr
    boundary_file =r'D:\D地图\广东省\广东市界镇界from zhu\广东市界R.shp'  
    variable_name="CO2"
    emis_paths=r"E:\CO2\data\emis\priori emission\MEIC_3km"
    emis_files=get_files(emis_paths)   
    
    model=model_attribute(emis_files[0])    
    x,y=model.get_xy_coordinates(show_lonlat=True)    
    dict_data={}
    unit=''            
    for index,emis_file in enumerate(emis_files):        
        ds = xr.open_dataset(emis_file, decode_times=False)
        if hasattr(ds[variable_name], 'units'):
            unit = ds[variable_name].units.strip()  
        file_name=os.path.basename(emis_file)  
        name=file_name.replace(f"EM_{file_name}_","") 
        if 'AG' in file_name:
            continue        
        get_multiple_data(dict_data,name,variable_name, x, y,np.mean(ds[variable_name][:,0,:,:],axis=0))
        if len(dict_data)==4:
            break
        
    fig=show_multi_maps(dict_data,show_original_grid=True,projection= model.projection,default_max_value=-1,default_min_value=-1,boundary_file=boundary_file,unit=unit,cmap='jet',showplot=False) 
    fig.savefig("dd.png")
     ###结束：绘制MEIC_3km数据的示例      
   
    print("Done")
   
   


import os
from esil import log
import sys
import chardet

def detect_encoding(file_path):
    with open(file_path, 'rb') as f:
        rawdata = f.read()
        result = chardet.detect(rawdata)
        return result['encoding']

def count_CSharp_code_lines(directory, file_extension=".cs"):
    '''
    @description: 统计指定目录下指定后缀名的文件的行数（不包括注释）
    @param directory: 指定目录
    @param file_extension: 指定后缀名
    @return: 行数
    '''
    total_lines = 0
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(file_extension):
                file_path = os.path.join(root, file)
                try:
                    encoding = detect_encoding(file_path)
                    with open(file_path, 'r', encoding=encoding) as f:
                        lines = f.readlines()
                        non_comment_lines = [line.strip() for line in lines if not line.strip().startswith("//") and not line.strip().startswith("/*") and not line.strip().endswith("*/") and line.strip() != ""]
                        total_lines += len(non_comment_lines)
                except Exception as e:
                    print(file_path)               
                    log.write(e, sys._getframe())
    return total_lines

def count_python_code_lines(directory, file_extension=".py"):
    '''
    @description: 统计指定目录下指定后缀名的文件的行数（不包括注释）
    @param directory: 指定目录
    @param file_extension: 指定后缀名
    @return: 行数
    '''
    total_lines = 0
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(file_extension):
                file_path = os.path.join(root, file)
                with open(file_path, 'r', encoding='utf-8') as f:
                    lines = f.readlines()
                    non_comment_lines = [line.strip() for line in lines if not line.strip().startswith("#") and line.strip() != ""]
                    total_lines += len(non_comment_lines)

    return total_lines

if __name__ == "__main__":
    project_directory = r"D:\Project\EPA\RSM-VAT\Code\ESIL.Kriging"  # 将路径替换为你的 C# 项目路径
    #project_directory = r"D:\Project\EPA\RSM-VAT\Code\RSM-VAT"  # 将路径替换为你的 C# 项目路径
    lines_of_code = count_CSharp_code_lines(project_directory)
    print(f"Total lines of code (excluding comments) in the C# project: {lines_of_code}")
    project_directory = r"D:\Project\EPA\RSM-VAT\Code\RSM-VAT\Script"  # 将路径替换为你的 C# 项目路径
    lines_of_code = count_python_code_lines(project_directory)
    print(f"Total lines of code (excluding comments) in the python scripts: {lines_of_code}")


import logging
from logging.handlers import TimedRotatingFileHandler
import os

# 配置日志记录器
def get_logger(log_file_name,identifier=None,when='midnight',interval=1,backupCount=0,suffix=None):
    '''
    @description: 获取日志记录器
    @param {str} log_file_name: 日志文件名，可以包含路径
    @param {str} identifier: 日志记录器的标识符，用于区分不同的日志记录器; 默认值为None，即使用调用者的文件名作为日志标识符   
    @param {str} when: 日志轮换频率，'S': 每秒, 'M': 每分钟,'H': 每小时,'D': 每天, 'W0' - 'W6': 每周的某一天（0为星期一，6为星期日）,'midnight': 每天午夜,默认值为midnight
    @param {int} interval: 日志轮换间隔，默认1
    @param {int} backupCount: 日志保留数量，默认0;保留的旧日志文件的数量，超过这个数量后会删除最旧的日志文件。
    @param {str} suffix: 日志文件后缀，默认值为None，即日期格式为"%Y-%m-%d_%H-%M-%S"
    @return: 日志记录器
    '''
    dir=os.path.dirname(log_file_name)
    if dir and not os.path.exists(dir):
        os.makedirs(dir)    
    if identifier is None:#使用调用者的文件名作为日志标识符 
        import inspect
        # 获取调用者的文件路径
        identifier =os.path.basename(inspect.stack()[1].filename).replace(".py","")    #inspect.stack()获取调用者的堆栈信息; 第二个元素是调用者的堆栈信息    
    logger = logging.getLogger(identifier)
    logger.setLevel(logging.INFO)    
    # 创建TimedRotatingFileHandler，每分钟生成一个带有日期时间戳的日志文件
    file_handler = logging.handlers.TimedRotatingFileHandler(log_file_name, when=when, interval=interval, backupCount=backupCount)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)
    if suffix is not None:
        file_handler.suffix = suffix
    logger.addHandler(file_handler)
    return logger

def write(e,currentFrame):
    # filename = sys._getframe().f_code.co_filename
    # function = sys._getframe().f_code.co_name
    # lineno = sys._getframe().f_lineno
    filename = currentFrame.f_code.co_filename
    function = currentFrame.f_code.co_name
    lineno = currentFrame.f_lineno
    exception = str(e)
    logErr(filename, function, lineno, exception)

def logErr(filename, function, lineno, e):
    import os   
    import time
    import logging
    current_path = os.path.abspath(__file__)
    parent_path = os.path.abspath(os.path.dirname(current_path) + os.path.sep + ".")
    today = time.strftime('%Y-%m-%d', time.localtime(time.time()))
    logName = today + ".log"
    if not os.path.exists("{}/Log/".format(parent_path)):
        os.makedirs("{}/Log/".format(parent_path))
    # if not os.path.exists("{}/Log/{}/".format(parent_path, today)):
    #     os.makedirs("{}/Log/{}/".format(parent_path, today))
    # if not os.path.exists("{}/Log/{}/{}".format(parent_path, today, logName)):
    #     reportFile = open("{}/Log/{}/{}".format(parent_path, today, logName), 'w')
    if not os.path.exists("{}/Log/{}".format(parent_path, logName)):
        reportFile = open("{}/Log/{}".format(parent_path, logName), 'w')
        reportFile.close()
    logger = logging.getLogger()
    handler = logging.FileHandler("{}/Log/{}".format(parent_path, logName), encoding='utf8')
    console = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(filename)s %(funcName)s %(levelname)s %(message)s')
    print(str(formatter))

    handler.setFormatter(formatter)  # 将log信息绑定到log文件上
    console.setFormatter(formatter)  # 将log信息绑定到控制台输出窗口
    logger.addHandler(handler)
    logger.addHandler(console)
    logger.setLevel(logging.INFO)  # Set log print level(设置日志打印级别)
    logging.error(filename + '   Function: ' + function + '  Line: ' + str(lineno) + '  Exception: ' + e+"\n")
    # print(e,'log.py')



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import BoundaryNorm

def plot_ekma_curve(x, y, values, colors_num=10, title='', x_title='VOCs', y_title='NOx', legend_title='O3', cmap='jet',
                    figsize=None, showplot=True):
    '''
    :param x: numpy 1D array
    :param y: numpy 1D array
    :param values: numpy 2D array,shape[len(x), len(y)]
    :param colors_num: 颜色个数
    :param title: 标题，default：''
    :param x_title: x轴标题,default:'VOCs'
    :param y_title: y轴标题,default:'NOx'
    :param legend_title: 图例标题,default:'O3'
    :param cmap: `matplotlib.colors.Colormap` or str or None, default: jet
        If a `.Colormap` instance, it will be returned. Otherwise, the name of
        a colormap known to Matplotlib, which will be resampled by *lut*. The
        value of None, means :rc:`image.cmap`.
    :param figsize: (float, float), default: :rc:`figure.figsize`
        Width, height in inches.
    :param showplot: 是否绘图，默认绘图；False将返回fig,用于保存
    :return: showplot=True，直接绘图，没有返回值。showplot=False,返回图表数据变量，可直接调用fig.savefig(output_file, dpi=600)进行保存图片。
    '''
    # 创建网格点
    # x_grid, NOx_grid = np.meshgrid(x, y)
    # 设定颜色分界点
    cmap_jet = plt.cm.get_cmap(cmap, colors_num)
    ticks = np.linspace(values.min(), values.max(), colors_num + 1)
    # boundaries = np.linspace(values.min(), values.max(), colors_num + 1)
    norm = BoundaryNorm(ticks, cmap_jet.N)
    # 绘制Ekman曲线
    fig = plt.figure(figsize=figsize)
    contour = plt.contourf(x, y, values, levels=ticks, cmap=cmap_jet, norm=norm)
    contour_lines = plt.contour(x, y, values, levels=contour.levels, colors='black',
                                linestyles='solid', linewidths=1)
    plt.clabel(contour_lines, inline=True, fontsize=8)
    cbar = plt.colorbar(contour, label=legend_title, ticks=ticks, boundaries=ticks)
    plt.xlabel(x_title)
    plt.ylabel(y_title)
    plt.title(title)
    # plt.grid(True)
    if showplot:
        # 显示图形
        plt.show()
    else:
        return fig

